"""System and user message construction helpers for Agent."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    Union,
)

from pydantic import BaseModel

if TYPE_CHECKING:
    from agno.agent.agent import Agent

from agno.agent._utils import convert_dependencies_to_string, convert_documents_to_string
from agno.filters import FilterExpr
from agno.media import Audio, File, Image, Video
from agno.models.message import Message, MessageReferences
from agno.models.response import ModelResponse
from agno.run import RunContext
from agno.run.agent import RunOutput
from agno.run.messages import RunMessages
from agno.session import AgentSession
from agno.tools.function import Function
from agno.utils.agent import (
    aexecute_instructions,
    aexecute_system_message,
    execute_instructions,
    execute_system_message,
)
from agno.utils.common import is_typed_dict
from agno.utils.log import log_debug, log_warning
from agno.utils.message import filter_tool_calls, get_text_from_message
from agno.utils.prompts import get_json_output_prompt, get_response_model_format_prompt
from agno.utils.timer import Timer


def _get_resolved_knowledge(agent: "Agent", run_context: Optional[RunContext] = None) -> Any:
    """Get the resolved knowledge, preferring run_context over agent.knowledge."""
    from agno.utils.callables import get_resolved_knowledge

    return get_resolved_knowledge(agent, run_context)


# ---------------------------------------------------------------------------
# Message formatting
# ---------------------------------------------------------------------------


def format_message_with_state_variables(
    agent: Agent,
    message: Any,
    run_context: Optional[RunContext] = None,
) -> Any:
    """Format a message with the session state variables from run_context."""
    import re
    import string
    from collections import ChainMap
    from copy import deepcopy

    if not isinstance(message, str):
        return message

    # Extract values from run_context
    session_state = run_context.session_state if run_context else None
    dependencies = run_context.dependencies if run_context else None
    metadata = run_context.metadata if run_context else None
    user_id = run_context.user_id if run_context else None

    # Should already be resolved and passed from run() method
    format_variables = ChainMap(
        session_state if session_state is not None else {},
        dependencies or {},
        metadata or {},
        {"user_id": user_id} if user_id is not None else {},
    )

    converted_msg = deepcopy(message)
    for var_name in format_variables.keys():
        # Only convert standalone {var_name} patterns, not nested ones
        pattern = r"\{" + re.escape(var_name) + r"\}"
        replacement = "${" + var_name + "}"
        converted_msg = re.sub(pattern, replacement, converted_msg)

    # Use Template to safely substitute variables
    template = string.Template(converted_msg)
    try:
        result = template.safe_substitute(format_variables)
        return result
    except Exception as e:
        log_warning(f"Template substitution failed: {e}")
        return message


# ---------------------------------------------------------------------------
# System message
# ---------------------------------------------------------------------------


def get_system_message(
    agent: Agent,
    session: AgentSession,
    run_context: Optional[RunContext] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    add_session_state_to_context: Optional[bool] = None,
) -> Optional[Message]:
    """Return the system message for the Agent.

    1. If the system_message is provided, use that.
    2. If build_context is False, return None.
    3. Build and return the default system message for the Agent.
    """

    # Extract values from run_context
    from agno.agent._init import set_culture_manager, set_memory_manager

    session_state = run_context.session_state if run_context else None
    user_id = run_context.user_id if run_context else None

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    # 1. If the system_message is provided, use that.
    if agent.system_message is not None:
        if isinstance(agent.system_message, Message):
            return agent.system_message

        sys_message_content: str = ""
        if isinstance(agent.system_message, str):
            sys_message_content = agent.system_message
        elif callable(agent.system_message):
            sys_message_content = execute_system_message(
                agent=agent, system_message=agent.system_message, session_state=session_state, run_context=run_context
            )
            if not isinstance(sys_message_content, str):
                raise Exception("system_message must return a string")

        if agent.resolve_in_context:
            sys_message_content = format_message_with_state_variables(
                agent,
                sys_message_content,
                run_context=run_context,
            )

        # type: ignore
        return Message(role=agent.system_message_role, content=sys_message_content)

    # 2. If build_context is False, return None.
    if not agent.build_context:
        return None

    if agent.model is None:
        raise Exception("model not set")

    # 3. Build and return the default system message for the Agent.
    # 3.1 Build the list of instructions for the system message
    instructions: List[str] = []
    if agent.instructions is not None:
        _instructions = agent.instructions
        if callable(agent.instructions):
            _instructions = execute_instructions(
                agent=agent, instructions=agent.instructions, session_state=session_state, run_context=run_context
            )

        if isinstance(_instructions, str):
            instructions.append(_instructions)
        elif isinstance(_instructions, list):
            instructions.extend(_instructions)

    # 3.1.1 Add instructions from the Model
    _model_instructions = agent.model.get_instructions_for_model(tools)
    if _model_instructions is not None:
        instructions.extend(_model_instructions)

    # 3.2 Build a list of additional information for the system message
    additional_information: List[str] = []
    # 3.2.1 Add instructions for using markdown
    if agent.markdown and output_schema is None:
        additional_information.append("Use markdown to format your answers.")
    # 3.2.2 Add the current datetime
    if agent.add_datetime_to_context:
        from datetime import datetime

        tz = None

        if agent.timezone_identifier:
            try:
                from zoneinfo import ZoneInfo

                tz = ZoneInfo(agent.timezone_identifier)
            except Exception:
                log_warning("Invalid timezone identifier")

        time = datetime.now(tz) if tz else datetime.now()

        if agent.datetime_format:
            formatted_time = time.strftime(agent.datetime_format)
        else:
            formatted_time = str(time)

        additional_information.append(f"The current time is {formatted_time}.")

    # 3.2.3 Add the current location
    if agent.add_location_to_context:
        from agno.utils.location import get_location

        location = get_location()
        if location:
            location_str = ", ".join(
                filter(
                    None,
                    [
                        location.get("city"),
                        location.get("region"),
                        location.get("country"),
                    ],
                )
            )
            if location_str:
                additional_information.append(f"Your approximate location is: {location_str}.")

    # 3.2.4 Add agent name if provided
    if agent.name is not None and agent.add_name_to_context:
        additional_information.append(f"Your name is: {agent.name}.")

    # 3.3 Build the default system message for the Agent.
    system_message_content: str = ""
    # 3.3.1 First add the Agent description if provided
    if agent.description is not None:
        system_message_content += f"{agent.description}\n"
    # 3.3.2 Then add the Agent role if provided
    if agent.role is not None:
        system_message_content += f"\n<your_role>\n{agent.role}\n</your_role>\n\n"
    # 3.3.3 Then add instructions for the Agent
    if len(instructions) > 0:
        if agent.use_instruction_tags:
            system_message_content += "<instructions>"
            if len(instructions) > 1:
                for _upi in instructions:
                    system_message_content += f"\n- {_upi}"
            else:
                system_message_content += "\n" + instructions[0]
            system_message_content += "\n</instructions>\n\n"
        else:
            if len(instructions) > 1:
                for _upi in instructions:
                    system_message_content += f"- {_upi}\n"
            else:
                system_message_content += instructions[0] + "\n\n"
    # 3.3.4 Add additional information
    if len(additional_information) > 0:
        system_message_content += "<additional_information>"
        for _ai in additional_information:
            system_message_content += f"\n- {_ai}"
        system_message_content += "\n</additional_information>\n\n"
    # 3.3.5 Then add instructions for the tools
    if agent._tool_instructions is not None:
        for _ti in agent._tool_instructions:
            system_message_content += f"{_ti}\n"

    # Format the system message with the session state variables
    if agent.resolve_in_context:
        system_message_content = format_message_with_state_variables(
            agent,
            system_message_content,
            run_context=run_context,
        )

    # 3.3.7 Then add the expected output
    if agent.expected_output is not None:
        system_message_content += f"<expected_output>\n{agent.expected_output.strip()}\n</expected_output>\n\n"
    # 3.3.8 Then add additional context
    if agent.additional_context is not None:
        system_message_content += f"{agent.additional_context}\n"
    # 3.3.8.1 Then add skills to the system prompt
    if agent.skills is not None:
        skills_snippet = agent.skills.get_system_prompt_snippet()
        if skills_snippet:
            system_message_content += f"\n{skills_snippet}\n"
    # 3.3.9 Then add memories to the system prompt
    if agent.add_memories_to_context:
        _memory_manager_not_set = False
        if not user_id:
            user_id = "default"
        if agent.memory_manager is None:
            set_memory_manager(agent)
            _memory_manager_not_set = True

        user_memories = agent.memory_manager.get_user_memories(user_id=user_id)  # type: ignore

        if user_memories and len(user_memories) > 0:
            system_message_content += "You have access to user info and preferences from previous interactions that you can use to personalize your response:\n\n"
            system_message_content += "<memories_from_previous_interactions>"
            for _memory in user_memories:  # type: ignore
                system_message_content += f"\n- {_memory.memory}"
            system_message_content += "\n</memories_from_previous_interactions>\n\n"
            system_message_content += (
                "Note: this information is from previous interactions and may be updated in this conversation. "
                "You should always prefer information from this conversation over the past memories.\n"
            )
        else:
            system_message_content += (
                "You have the capability to retain memories from previous interactions with the user, "
                "but have not had any interactions with the user yet.\n"
            )
        if _memory_manager_not_set:
            agent.memory_manager = None

        if agent.enable_agentic_memory:
            system_message_content += (
                "\n<updating_user_memories>\n"
                "- You have access to the `update_user_memory` tool that you can use to add new memories, update existing memories, delete memories, or clear all memories.\n"
                "- If the user's message includes information that should be captured as a memory, use the `update_user_memory` tool to update your memory database.\n"
                "- Memories should include details that could personalize ongoing interactions with the user.\n"
                "- Use this tool to add new memories or update existing memories that you identify in the conversation.\n"
                "- Use this tool if the user asks to update their memory, delete a memory, or clear all memories.\n"
                "- If you use the `update_user_memory` tool, remember to pass on the response to the user.\n"
                "</updating_user_memories>\n\n"
            )

    # 3.3.10 Then add cultural knowledge to the system prompt
    if agent.add_culture_to_context:
        _culture_manager_not_set = False
        if not agent.culture_manager:
            set_culture_manager(agent)
            _culture_manager_not_set = True

        cultural_knowledge = agent.culture_manager.get_all_knowledge()  # type: ignore

        if cultural_knowledge and len(cultural_knowledge) > 0:
            system_message_content += (
                "You have access to shared **Cultural Knowledge**, which provides context, norms, rules and guidance "
                "for your reasoning, communication, and decision-making. "
                "Cultural Knowledge represents the collective understanding, values, rules and practices that have "
                "emerged across agents and teams. It encodes collective experience — including preferred "
                "approaches, common patterns, lessons learned, and ethical guardrails.\n\n"
                "When performing any task:\n"
                "- **Reference Cultural Knowledge** to align with shared norms and best practices.\n"
                "- **Apply it contextually**, not mechanically — adapt principles to the current situation.\n"
                "- **Preserve consistency** with cultural values (tone, reasoning, and style) unless explicitly told otherwise.\n"
                "- **Extend it** when you discover new insights — your outputs may become future Cultural Knowledge.\n"
                "- **Clarify conflicts** if Cultural Knowledge appears to contradict explicit user instructions.\n\n"
                "Your goal is to act not only intelligently but also *culturally coherently* — reflecting the "
                "collective intelligence of the system.\n\n"
                "Below is the currently available Cultural Knowledge for this context:\n\n"
            )
            system_message_content += "<cultural_knowledge>"
            for _knowledge in cultural_knowledge:  # type: ignore
                system_message_content += "\n---"
                system_message_content += f"\nName: {_knowledge.name}"
                system_message_content += f"\nSummary: {_knowledge.summary}"
                system_message_content += f"\nContent: {_knowledge.content}"
            system_message_content += "\n</cultural_knowledge>\n"
        else:
            system_message_content += (
                "You have the capability to access shared **Cultural Knowledge**, which normally provides "
                "context, norms, and guidance for your behavior and reasoning. However, no cultural knowledge "
                "is currently available in this session.\n"
                "Proceed thoughtfully and document any useful insights you create — they may become future "
                "Cultural Knowledge for others.\n\n"
            )

        if _culture_manager_not_set:
            agent.culture_manager = None

        if agent.enable_agentic_culture:
            system_message_content += (
                "\n<contributing_to_culture>\n"
                "When you discover an insight, pattern, rule, or best practice that will help future agents, use the `create_or_update_cultural_knowledge` tool to add or update entries in the shared cultural knowledge.\n"
                "\n"
                "When to contribute:\n"
                "- You discover a reusable insight, pattern, rule, or best practice that will help future agents.\n"
                "- You correct or clarify an existing cultural entry.\n"
                "- You capture a guardrail, decision rationale, postmortem lesson, or example template.\n"
                "- You identify missing context that should persist across sessions or teams.\n"
                "\n"
                "Cultural knowledge should capture reusable insights, best practices, or contextual knowledge that transcends individual conversations.\n"
                "Mention your contribution to the user only if it is relevant to their request or they asked to be notified.\n"
                "</contributing_to_culture>\n\n"
            )

    # 3.3.11 Then add a summary of the interaction to the system prompt
    if agent.add_session_summary_to_context and session.summary is not None:
        system_message_content += "Here is a brief summary of your previous interactions:\n\n"
        system_message_content += "<summary_of_previous_interactions>\n"
        system_message_content += session.summary.summary
        system_message_content += "\n</summary_of_previous_interactions>\n\n"
        system_message_content += (
            "Note: this information is from previous interactions and may be outdated. "
            "You should ALWAYS prefer information from this conversation over the past summary.\n\n"
        )

    # 3.3.12 then add learnings to the system prompt
    if agent._learning is not None and agent.add_learnings_to_context:
        learning_context = agent._learning.build_context(
            user_id=user_id,
            session_id=session.session_id if session else None,
            agent_id=agent.id,
        )
        if learning_context:
            system_message_content += learning_context + "\n"

    # 3.3.13 then add search_knowledge instructions to the system prompt
    _resolved_knowledge = _get_resolved_knowledge(agent, run_context)
    if _resolved_knowledge is not None and agent.search_knowledge and agent.add_search_knowledge_instructions:
        build_context_fn = getattr(_resolved_knowledge, "build_context", None)
        if callable(build_context_fn):
            knowledge_context = build_context_fn(
                enable_agentic_filters=agent.enable_agentic_knowledge_filters,
            )
            if knowledge_context is not None:
                system_message_content += knowledge_context + "\n"

    # 3.3.14 Add the system message from the Model
    system_message_from_model = agent.model.get_system_message_for_model(tools)
    if system_message_from_model is not None:
        system_message_content += system_message_from_model

    # 3.3.15 Add the JSON output prompt if output_schema is provided and the model does not support native structured outputs or JSON schema outputs
    # or if use_json_mode is True
    if (
        output_schema is not None
        and agent.parser_model is None
        and not (
            (agent.model.supports_native_structured_outputs or agent.model.supports_json_schema_outputs)
            and (not agent.use_json_mode or agent.structured_outputs is True)
        )
    ):
        system_message_content += f"{get_json_output_prompt(output_schema)}"  # type: ignore

    # 3.3.16 Add the response model format prompt if output_schema is provided (Pydantic only)
    if output_schema is not None and agent.parser_model is not None and not isinstance(output_schema, dict):
        system_message_content += f"{get_response_model_format_prompt(output_schema)}"

    # 3.3.17 Add the session state to the system message
    if add_session_state_to_context and session_state is not None:
        system_message_content += f"\n<session_state>\n{session_state}\n</session_state>\n\n"

    # Return the system message
    return (
        Message(role=agent.system_message_role, content=system_message_content.strip())  # type: ignore
        if system_message_content
        else None
    )


async def aget_system_message(
    agent: Agent,
    session: AgentSession,
    run_context: Optional[RunContext] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    add_session_state_to_context: Optional[bool] = None,
) -> Optional[Message]:
    """Return the system message for the Agent.

    1. If the system_message is provided, use that.
    2. If build_context is False, return None.
    3. Build and return the default system message for the Agent.
    """

    # Extract values from run_context
    from agno.agent._init import has_async_db, set_culture_manager, set_memory_manager

    session_state = run_context.session_state if run_context else None
    user_id = run_context.user_id if run_context else None

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    # 1. If the system_message is provided, use that.
    if agent.system_message is not None:
        if isinstance(agent.system_message, Message):
            return agent.system_message

        sys_message_content: str = ""
        if isinstance(agent.system_message, str):
            sys_message_content = agent.system_message
        elif callable(agent.system_message):
            sys_message_content = await aexecute_system_message(
                agent=agent, system_message=agent.system_message, session_state=session_state, run_context=run_context
            )
            if not isinstance(sys_message_content, str):
                raise Exception("system_message must return a string")

        # Format the system message with the session state variables
        if agent.resolve_in_context:
            sys_message_content = format_message_with_state_variables(
                agent,
                sys_message_content,
                run_context=run_context,
            )

        # type: ignore
        return Message(role=agent.system_message_role, content=sys_message_content)

    # 2. If build_context is False, return None.
    if not agent.build_context:
        return None

    if agent.model is None:
        raise Exception("model not set")

    # 3. Build and return the default system message for the Agent.
    # 3.1 Build the list of instructions for the system message
    instructions: List[str] = []
    if agent.instructions is not None:
        _instructions = agent.instructions
        if callable(agent.instructions):
            _instructions = await aexecute_instructions(
                agent=agent, instructions=agent.instructions, session_state=session_state, run_context=run_context
            )

        if isinstance(_instructions, str):
            instructions.append(_instructions)
        elif isinstance(_instructions, list):
            instructions.extend(_instructions)

    # 3.1.1 Add instructions from the Model
    _model_instructions = agent.model.get_instructions_for_model(tools)
    if _model_instructions is not None:
        instructions.extend(_model_instructions)

    # 3.2 Build a list of additional information for the system message
    additional_information: List[str] = []
    # 3.2.1 Add instructions for using markdown
    if agent.markdown and output_schema is None:
        additional_information.append("Use markdown to format your answers.")
    # 3.2.2 Add the current datetime
    if agent.add_datetime_to_context:
        from datetime import datetime

        tz = None

        if agent.timezone_identifier:
            try:
                from zoneinfo import ZoneInfo

                tz = ZoneInfo(agent.timezone_identifier)
            except Exception:
                log_warning("Invalid timezone identifier")

        time = datetime.now(tz) if tz else datetime.now()

        if agent.datetime_format:
            formatted_time = time.strftime(agent.datetime_format)
        else:
            formatted_time = str(time)

        additional_information.append(f"The current time is {formatted_time}.")

    # 3.2.3 Add the current location
    if agent.add_location_to_context:
        from agno.utils.location import get_location

        location = get_location()
        if location:
            location_str = ", ".join(
                filter(
                    None,
                    [
                        location.get("city"),
                        location.get("region"),
                        location.get("country"),
                    ],
                )
            )
            if location_str:
                additional_information.append(f"Your approximate location is: {location_str}.")

    # 3.2.4 Add agent name if provided
    if agent.name is not None and agent.add_name_to_context:
        additional_information.append(f"Your name is: {agent.name}.")

    # 3.3 Build the default system message for the Agent.
    system_message_content: str = ""
    # 3.3.1 First add the Agent description if provided
    if agent.description is not None:
        system_message_content += f"{agent.description}\n"
    # 3.3.2 Then add the Agent role if provided
    if agent.role is not None:
        system_message_content += f"\n<your_role>\n{agent.role}\n</your_role>\n\n"
    # 3.3.3 Then add instructions for the Agent
    if len(instructions) > 0:
        if agent.use_instruction_tags:
            system_message_content += "<instructions>"
            if len(instructions) > 1:
                for _upi in instructions:
                    system_message_content += f"\n- {_upi}"
            else:
                system_message_content += "\n" + instructions[0]
            system_message_content += "\n</instructions>\n\n"
        else:
            if len(instructions) > 1:
                for _upi in instructions:
                    system_message_content += f"- {_upi}\n"
            else:
                system_message_content += instructions[0] + "\n\n"
    # 3.3.4 Add additional information
    if len(additional_information) > 0:
        system_message_content += "<additional_information>"
        for _ai in additional_information:
            system_message_content += f"\n- {_ai}"
        system_message_content += "\n</additional_information>\n\n"
    # 3.3.5 Then add instructions for the tools
    if agent._tool_instructions is not None:
        for _ti in agent._tool_instructions:
            system_message_content += f"{_ti}\n"

    # Format the system message with the session state variables
    if agent.resolve_in_context:
        system_message_content = format_message_with_state_variables(
            agent,
            system_message_content,
            run_context=run_context,
        )

    # 3.3.7 Then add the expected output
    if agent.expected_output is not None:
        system_message_content += f"<expected_output>\n{agent.expected_output.strip()}\n</expected_output>\n\n"
    # 3.3.8 Then add additional context
    if agent.additional_context is not None:
        system_message_content += f"{agent.additional_context}\n"
    # 3.3.8.1 Then add skills to the system prompt
    if agent.skills is not None:
        skills_snippet = agent.skills.get_system_prompt_snippet()
        if skills_snippet:
            system_message_content += f"\n{skills_snippet}\n"
    # 3.3.9 Then add memories to the system prompt
    if agent.add_memories_to_context:
        _memory_manager_not_set = False
        if not user_id:
            user_id = "default"
        if agent.memory_manager is None:
            set_memory_manager(agent)
            _memory_manager_not_set = True

        if has_async_db(agent):
            user_memories = await agent.memory_manager.aget_user_memories(user_id=user_id)  # type: ignore
        else:
            user_memories = agent.memory_manager.get_user_memories(user_id=user_id)  # type: ignore

        if user_memories and len(user_memories) > 0:
            system_message_content += "You have access to user info and preferences from previous interactions that you can use to personalize your response:\n\n"
            system_message_content += "<memories_from_previous_interactions>"
            for _memory in user_memories:  # type: ignore
                system_message_content += f"\n- {_memory.memory}"
            system_message_content += "\n</memories_from_previous_interactions>\n\n"
            system_message_content += (
                "Note: this information is from previous interactions and may be updated in this conversation. "
                "You should always prefer information from this conversation over the past memories.\n"
            )
        else:
            system_message_content += (
                "You have the capability to retain memories from previous interactions with the user, "
                "but have not had any interactions with the user yet.\n"
            )
        if _memory_manager_not_set:
            agent.memory_manager = None

        if agent.enable_agentic_memory:
            system_message_content += (
                "\n<updating_user_memories>\n"
                "- You have access to the `update_user_memory` tool that you can use to add new memories, update existing memories, delete memories, or clear all memories.\n"
                "- If the user's message includes information that should be captured as a memory, use the `update_user_memory` tool to update your memory database.\n"
                "- Memories should include details that could personalize ongoing interactions with the user.\n"
                "- Use this tool to add new memories or update existing memories that you identify in the conversation.\n"
                "- Use this tool if the user asks to update their memory, delete a memory, or clear all memories.\n"
                "- If you use the `update_user_memory` tool, remember to pass on the response to the user.\n"
                "</updating_user_memories>\n\n"
            )

    # 3.3.10 Then add cultural knowledge to the system prompt
    if agent.add_culture_to_context:
        _culture_manager_not_set = False
        if not agent.culture_manager:
            set_culture_manager(agent)
            _culture_manager_not_set = True

        cultural_knowledge = await agent.culture_manager.aget_all_knowledge()  # type: ignore

        if cultural_knowledge and len(cultural_knowledge) > 0:
            system_message_content += (
                "You have access to shared **Cultural Knowledge**, which provides context, norms, rules and guidance "
                "for your reasoning, communication, and decision-making.\n\n"
                "Cultural Knowledge represents the collective understanding, values, rules and practices that have "
                "emerged across agents and teams. It encodes collective experience — including preferred "
                "approaches, common patterns, lessons learned, and ethical guardrails.\n\n"
                "When performing any task:\n"
                "- **Reference Cultural Knowledge** to align with shared norms and best practices.\n"
                "- **Apply it contextually**, not mechanically — adapt principles to the current situation.\n"
                "- **Preserve consistency** with cultural values (tone, reasoning, and style) unless explicitly told otherwise.\n"
                "- **Extend it** when you discover new insights — your outputs may become future Cultural Knowledge.\n"
                "- **Clarify conflicts** if Cultural Knowledge appears to contradict explicit user instructions.\n\n"
                "Your goal is to act not only intelligently but also *culturally coherently* — reflecting the "
                "collective intelligence of the system.\n\n"
                "Below is the currently available Cultural Knowledge for this context:\n\n"
            )
            system_message_content += "<cultural_knowledge>"
            for _knowledge in cultural_knowledge:  # type: ignore
                system_message_content += "\n---"
                system_message_content += f"\nName: {_knowledge.name}"
                system_message_content += f"\nSummary: {_knowledge.summary}"
                system_message_content += f"\nContent: {_knowledge.content}"
            system_message_content += "\n</cultural_knowledge>\n"
        else:
            system_message_content += (
                "You have the capability to access shared **Cultural Knowledge**, which normally provides "
                "context, norms, and guidance for your behavior and reasoning. However, no cultural knowledge "
                "is currently available in this session.\n"
                "Proceed thoughtfully and document any useful insights you create — they may become future "
                "Cultural Knowledge for others.\n\n"
            )

        if _culture_manager_not_set:
            agent.culture_manager = None

        if agent.enable_agentic_culture:
            system_message_content += (
                "\n<contributing_to_culture>\n"
                "When you discover an insight, pattern, rule, or best practice that will help future agents, use the `create_or_update_cultural_knowledge` tool to add or update entries in the shared cultural knowledge.\n"
                "\n"
                "When to contribute:\n"
                "- You discover a reusable insight, pattern, rule, or best practice that will help future agents.\n"
                "- You correct or clarify an existing cultural entry.\n"
                "- You capture a guardrail, decision rationale, postmortem lesson, or example template.\n"
                "- You identify missing context that should persist across sessions or teams.\n"
                "\n"
                "Cultural knowledge should capture reusable insights, best practices, or contextual knowledge that transcends individual conversations.\n"
                "Mention your contribution to the user only if it is relevant to their request or they asked to be notified.\n"
                "</contributing_to_culture>\n\n"
            )

    # 3.3.11 Then add a summary of the interaction to the system prompt
    if agent.add_session_summary_to_context and session.summary is not None:
        system_message_content += "Here is a brief summary of your previous interactions:\n\n"
        system_message_content += "<summary_of_previous_interactions>\n"
        system_message_content += session.summary.summary
        system_message_content += "\n</summary_of_previous_interactions>\n\n"
        system_message_content += (
            "Note: this information is from previous interactions and may be outdated. "
            "You should ALWAYS prefer information from this conversation over the past summary.\n\n"
        )

    # 3.3.12 then add learnings to the system prompt
    if agent._learning is not None and agent.add_learnings_to_context:
        learning_context = await agent._learning.abuild_context(
            user_id=user_id,
            session_id=session.session_id if session else None,
            agent_id=agent.id,
        )
        if learning_context:
            system_message_content += learning_context + "\n"

    # 3.3.13 then add search_knowledge instructions to the system prompt
    _resolved_knowledge = _get_resolved_knowledge(agent, run_context)
    if _resolved_knowledge is not None and agent.search_knowledge and agent.add_search_knowledge_instructions:
        # Prefer async version if available for async databases
        abuild_context_fn = getattr(_resolved_knowledge, "abuild_context", None)
        build_context_fn = getattr(_resolved_knowledge, "build_context", None)
        if callable(abuild_context_fn):
            knowledge_context = await abuild_context_fn(
                enable_agentic_filters=agent.enable_agentic_knowledge_filters,
            )
            if knowledge_context is not None:
                system_message_content += knowledge_context + "\n"
        elif callable(build_context_fn):
            knowledge_context = build_context_fn(
                enable_agentic_filters=agent.enable_agentic_knowledge_filters,
            )
            if knowledge_context is not None:
                system_message_content += knowledge_context + "\n"

    # 3.3.14 Add the system message from the Model
    system_message_from_model = agent.model.get_system_message_for_model(tools)
    if system_message_from_model is not None:
        system_message_content += system_message_from_model

    # 3.3.15 Add the JSON output prompt if output_schema is provided and the model does not support native structured outputs or JSON schema outputs
    # or if use_json_mode is True
    if (
        output_schema is not None
        and agent.parser_model is None
        and not (
            (agent.model.supports_native_structured_outputs or agent.model.supports_json_schema_outputs)
            and (not agent.use_json_mode or agent.structured_outputs is True)
        )
    ):
        system_message_content += f"{get_json_output_prompt(output_schema)}"  # type: ignore

    # 3.3.16 Add the response model format prompt if output_schema is provided (Pydantic only)
    if output_schema is not None and agent.parser_model is not None and not isinstance(output_schema, dict):
        system_message_content += f"{get_response_model_format_prompt(output_schema)}"

    # 3.3.17 Add the session state to the system message
    if add_session_state_to_context and session_state is not None:
        system_message_content += get_formatted_session_state_for_system_message(agent, session_state)

    # Return the system message
    return (
        Message(role=agent.system_message_role, content=system_message_content.strip())  # type: ignore
        if system_message_content
        else None
    )


def get_formatted_session_state_for_system_message(agent: Agent, session_state: Dict[str, Any]) -> str:
    return f"\n<session_state>\n{session_state}\n</session_state>\n\n"


# ---------------------------------------------------------------------------
# User message
# ---------------------------------------------------------------------------


def get_user_message(
    agent: Agent,
    *,
    run_response: RunOutput,
    run_context: Optional[RunContext] = None,
    input: Optional[Union[str, List, Dict, Message, BaseModel, List[Message]]] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_dependencies_to_context: Optional[bool] = None,
    **kwargs: Any,
) -> Optional[Message]:
    """Return the user message for the Agent.

    1. If the user_message is provided, use that.
    2. If build_user_context is False or if the message is a list, return the message as is.
    3. Build the default user message for the Agent
    """
    # Extract values from run_context
    dependencies = run_context.dependencies if run_context else None
    knowledge_filters = run_context.knowledge_filters if run_context else None
    # Get references from the knowledge base to use in the user message
    references = None

    # 1. If build_user_context is False or message is a list, return the message as is.
    if not agent.build_user_context:
        return Message(
            role=agent.user_message_role or "user",
            content=input,  # type: ignore
            images=None if not agent.send_media_to_model else images,
            audio=None if not agent.send_media_to_model else audio,
            videos=None if not agent.send_media_to_model else videos,
            files=None if not agent.send_media_to_model else files,
            **kwargs,
        )
    # 2. Build the user message for the Agent
    elif input is None:
        # If we have any media, return a message with empty content
        if images is not None or audio is not None or videos is not None or files is not None:
            return Message(
                role=agent.user_message_role or "user",
                content="",
                images=None if not agent.send_media_to_model else images,
                audio=None if not agent.send_media_to_model else audio,
                videos=None if not agent.send_media_to_model else videos,
                files=None if not agent.send_media_to_model else files,
                **kwargs,
            )
        else:
            # If the input is None, return None
            return None

    else:
        # Handle list messages by converting to string
        if isinstance(input, list):
            # Convert list to string (join with newlines if all elements are strings)
            if all(isinstance(item, str) for item in input):
                message_content = "\n".join(input)  # type: ignore
            else:
                message_content = str(input)

            return Message(
                role=agent.user_message_role,
                content=message_content,
                images=None if not agent.send_media_to_model else images,
                audio=None if not agent.send_media_to_model else audio,
                videos=None if not agent.send_media_to_model else videos,
                files=None if not agent.send_media_to_model else files,
                **kwargs,
            )

        # If message is provided as a Message, use it directly
        elif isinstance(input, Message):
            return input
        # If message is provided as a dict, try to validate it as a Message
        elif isinstance(input, dict):
            try:
                return Message.model_validate(input)
            except Exception as e:
                log_warning(f"Failed to validate message: {e}")
                raise Exception(f"Failed to validate message: {e}")

        # If message is provided as a BaseModel, convert it to a Message
        elif isinstance(input, BaseModel):
            try:
                # Create a user message with the BaseModel content
                content = input.model_dump_json(indent=2, exclude_none=True)
                return Message(role=agent.user_message_role, content=content)
            except Exception as e:
                log_warning(f"Failed to convert BaseModel to message: {e}")
                raise Exception(f"Failed to convert BaseModel to message: {e}")
        else:
            user_msg_content = input
            if agent.add_knowledge_to_context:
                if isinstance(input, str):
                    user_msg_content = input
                elif callable(input):
                    user_msg_content = input(agent=agent)
                else:
                    raise Exception("message must be a string or a callable when add_references is True")

                try:
                    retrieval_timer = Timer()
                    retrieval_timer.start()
                    docs_from_knowledge = get_relevant_docs_from_knowledge(
                        agent, query=user_msg_content, filters=knowledge_filters, run_context=run_context, **kwargs
                    )
                    if docs_from_knowledge is not None:
                        references = MessageReferences(
                            query=user_msg_content,
                            references=docs_from_knowledge,
                            time=round(retrieval_timer.elapsed, 4),
                        )
                        # Add the references to the run_response
                        if run_response.references is None:
                            run_response.references = []
                        run_response.references.append(references)
                    retrieval_timer.stop()
                    log_debug(f"Time to get references: {retrieval_timer.elapsed:.4f}s")
                except Exception as e:
                    log_warning(f"Failed to get references: {e}")

            if agent.resolve_in_context:
                user_msg_content = format_message_with_state_variables(
                    agent,
                    user_msg_content,
                    run_context=run_context,
                )

            # Convert to string for concatenation operations
            user_msg_content_str = get_text_from_message(user_msg_content) if user_msg_content is not None else ""

            # 4.1 Add knowledge references to user message
            if (
                agent.add_knowledge_to_context
                and references is not None
                and references.references is not None
                and len(references.references) > 0
            ):
                user_msg_content_str += "\n\nUse the following references from the knowledge base if it helps:\n"
                user_msg_content_str += "<references>\n"
                user_msg_content_str += convert_documents_to_string(agent, references.references) + "\n"
                user_msg_content_str += "</references>"
            # 4.2 Add context to user message
            if add_dependencies_to_context and dependencies is not None:
                user_msg_content_str += "\n\n<additional context>\n"
                user_msg_content_str += convert_dependencies_to_string(agent, dependencies) + "\n"
                user_msg_content_str += "</additional context>"

            # Use the string version for the final content
            user_msg_content = user_msg_content_str

            # Return the user message
            return Message(
                role=agent.user_message_role,
                content=user_msg_content,
                audio=None if not agent.send_media_to_model else audio,
                images=None if not agent.send_media_to_model else images,
                videos=None if not agent.send_media_to_model else videos,
                files=None if not agent.send_media_to_model else files,
                **kwargs,
            )


async def aget_user_message(
    agent: Agent,
    *,
    run_response: RunOutput,
    run_context: Optional[RunContext] = None,
    input: Optional[Union[str, List, Dict, Message, BaseModel, List[Message]]] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_dependencies_to_context: Optional[bool] = None,
    **kwargs: Any,
) -> Optional[Message]:
    """Return the user message for the Agent (async version).

    1. If the user_message is provided, use that.
    2. If build_user_context is False or if the message is a list, return the message as is.
    3. Build the default user message for the Agent
    """
    # Extract values from run_context
    dependencies = run_context.dependencies if run_context else None
    knowledge_filters = run_context.knowledge_filters if run_context else None
    # Get references from the knowledge base to use in the user message
    references = None

    # 1. If build_user_context is False or message is a list, return the message as is.
    if not agent.build_user_context:
        return Message(
            role=agent.user_message_role or "user",
            content=input,  # type: ignore
            images=None if not agent.send_media_to_model else images,
            audio=None if not agent.send_media_to_model else audio,
            videos=None if not agent.send_media_to_model else videos,
            files=None if not agent.send_media_to_model else files,
            **kwargs,
        )
    # 2. Build the user message for the Agent
    elif input is None:
        # If we have any media, return a message with empty content
        if images is not None or audio is not None or videos is not None or files is not None:
            return Message(
                role=agent.user_message_role or "user",
                content="",
                images=None if not agent.send_media_to_model else images,
                audio=None if not agent.send_media_to_model else audio,
                videos=None if not agent.send_media_to_model else videos,
                files=None if not agent.send_media_to_model else files,
                **kwargs,
            )
        else:
            # If the input is None, return None
            return None

    else:
        # Handle list messages by converting to string
        if isinstance(input, list):
            # Convert list to string (join with newlines if all elements are strings)
            if all(isinstance(item, str) for item in input):
                message_content = "\n".join(input)  # type: ignore
            else:
                message_content = str(input)

            return Message(
                role=agent.user_message_role,
                content=message_content,
                images=None if not agent.send_media_to_model else images,
                audio=None if not agent.send_media_to_model else audio,
                videos=None if not agent.send_media_to_model else videos,
                files=None if not agent.send_media_to_model else files,
                **kwargs,
            )

        # If message is provided as a Message, use it directly
        elif isinstance(input, Message):
            return input
        # If message is provided as a dict, try to validate it as a Message
        elif isinstance(input, dict):
            try:
                return Message.model_validate(input)
            except Exception as e:
                log_warning(f"Failed to validate message: {e}")
                raise Exception(f"Failed to validate message: {e}")

        # If message is provided as a BaseModel, convert it to a Message
        elif isinstance(input, BaseModel):
            try:
                # Create a user message with the BaseModel content
                content = input.model_dump_json(indent=2, exclude_none=True)
                return Message(role=agent.user_message_role, content=content)
            except Exception as e:
                log_warning(f"Failed to convert BaseModel to message: {e}")
                raise Exception(f"Failed to convert BaseModel to message: {e}")
        else:
            user_msg_content = input
            if agent.add_knowledge_to_context:
                if isinstance(input, str):
                    user_msg_content = input
                elif callable(input):
                    user_msg_content = input(agent=agent)
                else:
                    raise Exception("message must be a string or a callable when add_references is True")

                try:
                    retrieval_timer = Timer()
                    retrieval_timer.start()
                    docs_from_knowledge = await aget_relevant_docs_from_knowledge(
                        agent, query=user_msg_content, filters=knowledge_filters, run_context=run_context, **kwargs
                    )
                    if docs_from_knowledge is not None:
                        references = MessageReferences(
                            query=user_msg_content,
                            references=docs_from_knowledge,
                            time=round(retrieval_timer.elapsed, 4),
                        )
                        # Add the references to the run_response
                        if run_response.references is None:
                            run_response.references = []
                        run_response.references.append(references)
                    retrieval_timer.stop()
                    log_debug(f"Time to get references: {retrieval_timer.elapsed:.4f}s")
                except Exception as e:
                    log_warning(f"Failed to get references: {e}")

            if agent.resolve_in_context:
                user_msg_content = format_message_with_state_variables(
                    agent,
                    user_msg_content,
                    run_context=run_context,
                )

            # Convert to string for concatenation operations
            user_msg_content_str = get_text_from_message(user_msg_content) if user_msg_content is not None else ""

            # 4.1 Add knowledge references to user message
            if (
                agent.add_knowledge_to_context
                and references is not None
                and references.references is not None
                and len(references.references) > 0
            ):
                user_msg_content_str += "\n\nUse the following references from the knowledge base if it helps:\n"
                user_msg_content_str += "<references>\n"
                user_msg_content_str += convert_documents_to_string(agent, references.references) + "\n"
                user_msg_content_str += "</references>"
            # 4.2 Add context to user message
            if add_dependencies_to_context and dependencies is not None:
                user_msg_content_str += "\n\n<additional context>\n"
                user_msg_content_str += convert_dependencies_to_string(agent, dependencies) + "\n"
                user_msg_content_str += "</additional context>"

            # Use the string version for the final content
            user_msg_content = user_msg_content_str

            # Return the user message
            return Message(
                role=agent.user_message_role,
                content=user_msg_content,
                audio=None if not agent.send_media_to_model else audio,
                images=None if not agent.send_media_to_model else images,
                videos=None if not agent.send_media_to_model else videos,
                files=None if not agent.send_media_to_model else files,
                **kwargs,
            )


# ---------------------------------------------------------------------------
# Run messages
# ---------------------------------------------------------------------------


def get_run_messages(
    agent: Agent,
    *,
    run_response: RunOutput,
    run_context: RunContext,
    input: Union[str, List, Dict, Message, BaseModel, List[Message]],
    session: AgentSession,
    user_id: Optional[str] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_history_to_context: Optional[bool] = None,
    add_dependencies_to_context: Optional[bool] = None,
    add_session_state_to_context: Optional[bool] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    **kwargs: Any,
) -> RunMessages:
    """This function returns a RunMessages object with the following attributes:
        - system_message: The system message for this run
        - user_message: The user message for this run
        - messages: List of messages to send to the model

    To build the RunMessages object:
    1. Add system message to run_messages
    2. Add extra messages to run_messages if provided
    3. Add history to run_messages
    4. Add user message to run_messages (if input is single content)
    5. Add input messages to run_messages if provided (if input is List[Message])

    Returns:
        RunMessages object with the following attributes:
            - system_message: The system message for this run
            - user_message: The user message for this run
            - messages: List of all messages to send to the model

    Typical usage:
    run_messages = get_run_messages(
        agent, input=input, session_id=session_id, user_id=user_id, audio=audio, images=images, videos=videos, files=files, **kwargs
    )
    """

    # Initialize the RunMessages object (no media here - that's in RunInput now)
    run_messages = RunMessages()

    # 1. Add system message to run_messages
    system_message = get_system_message(
        agent,
        session=session,
        run_context=run_context,
        tools=tools,
        add_session_state_to_context=add_session_state_to_context,
    )
    if system_message is not None:
        run_messages.system_message = system_message
        run_messages.messages.append(system_message)

    # 2. Add extra messages to run_messages if provided
    if agent.additional_input is not None:
        messages_to_add_to_run_response: List[Message] = []
        if run_messages.extra_messages is None:
            run_messages.extra_messages = []

        for _m in agent.additional_input:
            if isinstance(_m, Message):
                messages_to_add_to_run_response.append(_m)
                run_messages.messages.append(_m)
                run_messages.extra_messages.append(_m)
            elif isinstance(_m, dict):
                try:
                    _m_parsed = Message.model_validate(_m)
                    messages_to_add_to_run_response.append(_m_parsed)
                    run_messages.messages.append(_m_parsed)
                    run_messages.extra_messages.append(_m_parsed)
                except Exception as e:
                    log_warning(f"Failed to validate message: {e}")
        # Add the extra messages to the run_response
        if len(messages_to_add_to_run_response) > 0:
            log_debug(f"Adding {len(messages_to_add_to_run_response)} extra messages")
            if run_response.additional_input is None:
                run_response.additional_input = messages_to_add_to_run_response
            else:
                run_response.additional_input.extend(messages_to_add_to_run_response)

    # 3. Add history to run_messages
    if add_history_to_context:
        from copy import deepcopy

        # Only skip messages from history when system_message_role is NOT a standard conversation role.
        # Standard conversation roles ("user", "assistant", "tool") should never be filtered
        # to preserve conversation continuity.
        skip_role = (
            agent.system_message_role if agent.system_message_role not in ["user", "assistant", "tool"] else None
        )

        history: List[Message] = session.get_messages(
            last_n_runs=agent.num_history_runs,
            limit=agent.num_history_messages,
            skip_roles=[skip_role] if skip_role else None,
            agent_id=agent.id if agent.team_id is not None else None,
        )

        if len(history) > 0:
            # Create a deep copy of the history messages to avoid modifying the original messages
            history_copy = [deepcopy(msg) for msg in history]

            # Tag each message as coming from history
            for _msg in history_copy:
                _msg.from_history = True

            # Filter tool calls from history if limit is set (before adding to run_messages)
            if agent.max_tool_calls_from_history is not None:
                filter_tool_calls(history_copy, agent.max_tool_calls_from_history)

            log_debug(f"Adding {len(history_copy)} messages from history")

            run_messages.messages += history_copy

    # 4. Add user message to run_messages
    user_message: Optional[Message] = None

    # 4.1 Build user message if input is None, str or list and not a list of Message/dict objects
    if (
        input is None
        or isinstance(input, str)
        or (
            isinstance(input, list)
            and not (
                len(input) > 0
                and (isinstance(input[0], Message) or (isinstance(input[0], dict) and "role" in input[0]))
            )
        )
    ):
        user_message = get_user_message(
            agent,
            run_response=run_response,
            run_context=run_context,
            input=input,
            audio=audio,
            images=images,
            videos=videos,
            files=files,
            add_dependencies_to_context=add_dependencies_to_context,
            **kwargs,
        )

    # 4.2 If input is provided as a Message, use it directly
    elif isinstance(input, Message):
        user_message = input

    # 4.3 If input is provided as a dict, try to validate it as a Message
    elif isinstance(input, dict):
        try:
            if agent.input_schema and is_typed_dict(agent.input_schema):
                import json

                content = json.dumps(input, indent=2, ensure_ascii=False)
                user_message = Message(role=agent.user_message_role, content=content)
            else:
                user_message = Message.model_validate(input)
        except Exception as e:
            log_warning(f"Failed to validate message: {e}")

    # 4.4 If input is provided as a BaseModel, convert it to a Message
    elif isinstance(input, BaseModel):
        try:
            # Create a user message with the BaseModel content
            content = input.model_dump_json(indent=2, exclude_none=True)
            user_message = Message(role=agent.user_message_role, content=content)
        except Exception as e:
            log_warning(f"Failed to convert BaseModel to message: {e}")

    # 5. Add input messages to run_messages if provided (List[Message] or List[Dict])
    if (
        isinstance(input, list)
        and len(input) > 0
        and (isinstance(input[0], Message) or (isinstance(input[0], dict) and "role" in input[0]))
    ):
        for _m in input:
            if isinstance(_m, Message):
                run_messages.messages.append(_m)
                if run_messages.extra_messages is None:
                    run_messages.extra_messages = []
                run_messages.extra_messages.append(_m)
            elif isinstance(_m, dict):
                try:
                    msg = Message.model_validate(_m)
                    run_messages.messages.append(msg)
                    if run_messages.extra_messages is None:
                        run_messages.extra_messages = []
                    run_messages.extra_messages.append(msg)
                except Exception as e:
                    log_warning(f"Failed to validate message: {e}")

    # Add user message to run_messages
    if user_message is not None:
        run_messages.user_message = user_message
        run_messages.messages.append(user_message)

    # Set messages on run_context so tool hooks can access the current message history
    run_context.messages = run_messages.messages

    return run_messages


async def aget_run_messages(
    agent: Agent,
    *,
    run_response: RunOutput,
    run_context: RunContext,
    input: Union[str, List, Dict, Message, BaseModel, List[Message]],
    session: AgentSession,
    user_id: Optional[str] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_history_to_context: Optional[bool] = None,
    add_dependencies_to_context: Optional[bool] = None,
    add_session_state_to_context: Optional[bool] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    **kwargs: Any,
) -> RunMessages:
    """This function returns a RunMessages object with the following attributes:
        - system_message: The system message for this run
        - user_message: The user message for this run
        - messages: List of messages to send to the model

    To build the RunMessages object:
    1. Add system message to run_messages
    2. Add extra messages to run_messages if provided
    3. Add history to run_messages
    4. Add user message to run_messages (if input is single content)
    5. Add input messages to run_messages if provided (if input is List[Message])

    Returns:
        RunMessages object with the following attributes:
            - system_message: The system message for this run
            - user_message: The user message for this run
            - messages: List of all messages to send to the model

    Typical usage:
    run_messages = await aget_run_messages(
        agent, input=input, session_id=session_id, user_id=user_id, audio=audio, images=images, videos=videos, files=files, **kwargs
    )
    """

    # Initialize the RunMessages object (no media here - that's in RunInput now)
    run_messages = RunMessages()

    # 1. Add system message to run_messages
    system_message = await aget_system_message(
        agent,
        session=session,
        run_context=run_context,
        tools=tools,
        add_session_state_to_context=add_session_state_to_context,
    )
    if system_message is not None:
        run_messages.system_message = system_message
        run_messages.messages.append(system_message)

    # 2. Add extra messages to run_messages if provided
    if agent.additional_input is not None:
        messages_to_add_to_run_response: List[Message] = []
        if run_messages.extra_messages is None:
            run_messages.extra_messages = []

        for _m in agent.additional_input:
            if isinstance(_m, Message):
                messages_to_add_to_run_response.append(_m)
                run_messages.messages.append(_m)
                run_messages.extra_messages.append(_m)
            elif isinstance(_m, dict):
                try:
                    _m_parsed = Message.model_validate(_m)
                    messages_to_add_to_run_response.append(_m_parsed)
                    run_messages.messages.append(_m_parsed)
                    run_messages.extra_messages.append(_m_parsed)
                except Exception as e:
                    log_warning(f"Failed to validate message: {e}")
        # Add the extra messages to the run_response
        if len(messages_to_add_to_run_response) > 0:
            log_debug(f"Adding {len(messages_to_add_to_run_response)} extra messages")
            if run_response.additional_input is None:
                run_response.additional_input = messages_to_add_to_run_response
            else:
                run_response.additional_input.extend(messages_to_add_to_run_response)

    # 3. Add history to run_messages
    if add_history_to_context:
        from copy import deepcopy

        # Only skip messages from history when system_message_role is NOT a standard conversation role.
        # Standard conversation roles ("user", "assistant", "tool") should never be filtered
        # to preserve conversation continuity.
        skip_role = (
            agent.system_message_role if agent.system_message_role not in ["user", "assistant", "tool"] else None
        )

        history: List[Message] = session.get_messages(
            last_n_runs=agent.num_history_runs,
            limit=agent.num_history_messages,
            skip_roles=[skip_role] if skip_role else None,
            agent_id=agent.id if agent.team_id is not None else None,
        )

        if len(history) > 0:
            # Create a deep copy of the history messages to avoid modifying the original messages
            history_copy = [deepcopy(msg) for msg in history]

            # Tag each message as coming from history
            for _msg in history_copy:
                _msg.from_history = True

            # Filter tool calls from history if limit is set (before adding to run_messages)
            if agent.max_tool_calls_from_history is not None:
                filter_tool_calls(history_copy, agent.max_tool_calls_from_history)

            log_debug(f"Adding {len(history_copy)} messages from history")

            run_messages.messages += history_copy

    # 4. Add user message to run_messages
    user_message: Optional[Message] = None

    # 4.1 Build user message if input is None, str or list and not a list of Message/dict objects
    if (
        input is None
        or isinstance(input, str)
        or (
            isinstance(input, list)
            and not (
                len(input) > 0
                and (isinstance(input[0], Message) or (isinstance(input[0], dict) and "role" in input[0]))
            )
        )
    ):
        user_message = await aget_user_message(
            agent,
            run_response=run_response,
            run_context=run_context,
            input=input,
            audio=audio,
            images=images,
            videos=videos,
            files=files,
            add_dependencies_to_context=add_dependencies_to_context,
            **kwargs,
        )

    # 4.2 If input is provided as a Message, use it directly
    elif isinstance(input, Message):
        user_message = input

    # 4.3 If input is provided as a dict, try to validate it as a Message
    elif isinstance(input, dict):
        try:
            if agent.input_schema and is_typed_dict(agent.input_schema):
                import json

                content = json.dumps(input, indent=2, ensure_ascii=False)
                user_message = Message(role=agent.user_message_role, content=content)
            else:
                user_message = Message.model_validate(input)
        except Exception as e:
            log_warning(f"Failed to validate message: {e}")

    # 4.4 If input is provided as a BaseModel, convert it to a Message
    elif isinstance(input, BaseModel):
        try:
            # Create a user message with the BaseModel content
            content = input.model_dump_json(indent=2, exclude_none=True)
            user_message = Message(role=agent.user_message_role, content=content)
        except Exception as e:
            log_warning(f"Failed to convert BaseModel to message: {e}")

    # 5. Add input messages to run_messages if provided (List[Message] or List[Dict])
    if (
        isinstance(input, list)
        and len(input) > 0
        and (isinstance(input[0], Message) or (isinstance(input[0], dict) and "role" in input[0]))
    ):
        for _m in input:
            if isinstance(_m, Message):
                run_messages.messages.append(_m)
                if run_messages.extra_messages is None:
                    run_messages.extra_messages = []
                run_messages.extra_messages.append(_m)
            elif isinstance(_m, dict):
                try:
                    msg = Message.model_validate(_m)
                    run_messages.messages.append(msg)
                    if run_messages.extra_messages is None:
                        run_messages.extra_messages = []
                    run_messages.extra_messages.append(msg)
                except Exception as e:
                    log_warning(f"Failed to validate message: {e}")

    # Add user message to run_messages
    if user_message is not None:
        run_messages.user_message = user_message
        run_messages.messages.append(user_message)

    # Set messages on run_context so tool hooks can access the current message history
    run_context.messages = run_messages.messages

    return run_messages


def get_continue_run_messages(
    agent: Agent,
    input: List[Message],
    session: Optional[AgentSession] = None,
    add_history_to_context: Optional[bool] = None,
    run_context: Optional[RunContext] = None,
) -> RunMessages:
    """This function returns a RunMessages object with the following attributes:
        - system_message: The system message for this run
        - user_message: The user message for this run
        - messages: List of messages to send to the model

    It continues from a previous run and completes a tool call that was paused.
    """

    # Initialize the RunMessages object
    run_messages = RunMessages()

    # Add history to run_messages (optional)
    if add_history_to_context is None:
        add_history_to_context = agent.add_history_to_context

    # Extract most recent user message from messages as the original user message
    user_message = None
    for msg in reversed(input):
        if msg.role == agent.user_message_role:
            user_message = msg
            break

    # Extract system message from messages
    system_message = None
    for msg in input:
        if msg.role == agent.system_message_role:
            system_message = msg
            break

    run_messages.system_message = system_message
    run_messages.user_message = user_message

    # Skip re-fetching history if input already contains it (run_response path).
    input_has_history = any(msg.from_history for msg in input)

    # Build messages in the correct order:
    # 1. System message first
    # 2. History messages (from previous runs, only if not already in input)
    # 3. Remaining input messages (current run's messages, excluding the system message)

    # 1. Add the system message first
    if system_message is not None:
        run_messages.messages.append(system_message)

    # 2. Add history messages if not already present in input
    if add_history_to_context and session is not None and not input_has_history:
        from copy import deepcopy

        # Only skip messages from history when system_message_role is NOT a standard conversation role.
        # Standard conversation roles ("user", "assistant", "tool") should never be filtered
        # to preserve conversation continuity.
        skip_role = (
            agent.system_message_role if agent.system_message_role not in ["user", "assistant", "tool"] else None
        )

        history: List[Message] = session.get_messages(
            last_n_runs=agent.num_history_runs,
            limit=agent.num_history_messages,
            skip_roles=[skip_role] if skip_role else None,
            agent_id=agent.id if agent.team_id is not None else None,
        )

        if len(history) > 0:
            # Create a deep copy of the history messages to avoid modifying the original messages
            history_copy = [deepcopy(msg) for msg in history]

            # Tag each message as coming from history
            for _msg in history_copy:
                _msg.from_history = True

            # Filter tool calls from history if limit is set (before adding to run_messages)
            if agent.max_tool_calls_from_history is not None:
                filter_tool_calls(history_copy, agent.max_tool_calls_from_history)

            log_debug(f"Adding {len(history_copy)} messages from history")
            run_messages.messages += history_copy

    # 3. Add the remaining input messages (skip the system message to avoid duplication)
    for msg in input:
        if msg is not system_message:
            run_messages.messages.append(msg)

    # Set messages on run_context so tool hooks can access the current message history
    if run_context is not None:
        run_context.messages = run_messages.messages

    return run_messages


# ---------------------------------------------------------------------------
# Parser / output model messages
# ---------------------------------------------------------------------------


def get_messages_for_parser_model(
    agent: Agent,
    model_response: ModelResponse,
    response_format: Optional[Union[Dict, Type[BaseModel]]],
    run_context: Optional[RunContext] = None,
) -> List[Message]:
    """Get the messages for the parser model."""
    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    system_content = (
        agent.parser_model_prompt
        if agent.parser_model_prompt is not None
        else "You are tasked with creating a structured output from the provided user message."
    )

    if response_format == {"type": "json_object"} and output_schema is not None:
        system_content += f"{get_json_output_prompt(output_schema)}"  # type: ignore

    return [
        Message(role="system", content=system_content),
        Message(role="user", content=model_response.content),
    ]


def get_messages_for_parser_model_stream(
    agent: Agent,
    run_response: RunOutput,
    response_format: Optional[Union[Dict, Type[BaseModel]]],
    run_context: Optional[RunContext] = None,
) -> List[Message]:
    """Get the messages for the parser model."""
    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    system_content = (
        agent.parser_model_prompt
        if agent.parser_model_prompt is not None
        else "You are tasked with creating a structured output from the provided data."
    )

    if response_format == {"type": "json_object"} and output_schema is not None:
        system_content += f"{get_json_output_prompt(output_schema)}"  # type: ignore

    return [
        Message(role="system", content=system_content),
        Message(role="user", content=run_response.content),
    ]


def get_messages_for_output_model(agent: Agent, messages: List[Message]) -> List[Message]:
    """Get the messages for the output model."""

    if agent.output_model_prompt is not None:
        system_message_exists = False
        for message in messages:
            if message.role == "system":
                system_message_exists = True
                message.content = agent.output_model_prompt
                break
        if not system_message_exists:
            messages.insert(0, Message(role="system", content=agent.output_model_prompt))

    # Remove the last assistant message from the messages list
    messages.pop(-1)

    return messages


# ---------------------------------------------------------------------------
# Knowledge retrieval
# ---------------------------------------------------------------------------


def get_relevant_docs_from_knowledge(
    agent: Agent,
    query: str,
    num_documents: Optional[int] = None,
    filters: Optional[Union[Dict[str, Any], List[FilterExpr]]] = None,
    validate_filters: bool = False,
    run_context: Optional[RunContext] = None,
    **kwargs: Any,
) -> Optional[List[Union[Dict[str, Any], str]]]:
    """Get relevant docs from the knowledge base to answer a query.

    Args:
        agent: The Agent instance.
        query (str): The query to search for.
        num_documents (Optional[int]): Number of documents to return.
        filters (Optional[Dict[str, Any]]): Filters to apply to the search.
        validate_filters (bool): Whether to validate the filters against known valid filter keys.
        run_context (Optional[RunContext]): Runtime context containing dependencies and other context.
        **kwargs: Additional keyword arguments.

    Returns:
        Optional[List[Dict[str, Any]]]: List of relevant document dicts.
    """
    from agno.knowledge.document import Document

    # Extract dependencies from run_context if available
    dependencies = run_context.dependencies if run_context else None

    resolved_knowledge = _get_resolved_knowledge(agent, run_context)

    if num_documents is None and resolved_knowledge is not None:
        num_documents = getattr(resolved_knowledge, "max_results", None)
    # Validate the filters against known valid filter keys
    if resolved_knowledge is not None and filters is not None:
        if validate_filters:
            valid_filters, invalid_keys = resolved_knowledge.validate_filters(filters)  # type: ignore

            # Warn about invalid filter keys
            if invalid_keys:
                # type: ignore
                log_warning(f"Invalid filter keys provided: {invalid_keys}. These filters will be ignored.")

                # Only use valid filters
                filters = valid_filters
                if not filters:
                    log_warning("No valid filters remain after validation. Search will proceed without filters.")

            if invalid_keys == [] and valid_filters == {}:
                log_warning("No valid filters provided. Search will proceed without filters.")
                filters = None

    if agent.knowledge_retriever is not None and callable(agent.knowledge_retriever):
        from inspect import signature

        try:
            sig = signature(agent.knowledge_retriever)
            knowledge_retriever_kwargs: Dict[str, Any] = {}
            if "agent" in sig.parameters:
                knowledge_retriever_kwargs = {"agent": agent}
            if "filters" in sig.parameters:
                knowledge_retriever_kwargs["filters"] = filters
            if "run_context" in sig.parameters:
                knowledge_retriever_kwargs["run_context"] = run_context
            elif "dependencies" in sig.parameters:
                # Backward compatibility: support dependencies parameter
                knowledge_retriever_kwargs["dependencies"] = dependencies
            knowledge_retriever_kwargs.update({"query": query, "num_documents": num_documents, **kwargs})
            return agent.knowledge_retriever(**knowledge_retriever_kwargs)
        except Exception as e:
            log_warning(f"Knowledge retriever failed: {e}")
            raise e

    # Use knowledge protocol's retrieve method
    try:
        if resolved_knowledge is None:
            return None

        # Use protocol retrieve() method if available
        retrieve_fn = getattr(resolved_knowledge, "retrieve", None)
        if not callable(retrieve_fn):
            log_debug("Knowledge does not implement retrieve()")
            return None

        if num_documents is None:
            num_documents = getattr(resolved_knowledge, "max_results", 10)

        log_debug(f"Retrieving from knowledge base with filters: {filters}")
        relevant_docs: List[Document] = retrieve_fn(query=query, max_results=num_documents, filters=filters)

        if not relevant_docs or len(relevant_docs) == 0:
            log_debug("No relevant documents found for query")
            return None

        return [doc.to_dict() for doc in relevant_docs]
    except Exception as e:
        log_warning(f"Error retrieving from knowledge base: {e}")
        raise e


async def aget_relevant_docs_from_knowledge(
    agent: Agent,
    query: str,
    num_documents: Optional[int] = None,
    filters: Optional[Union[Dict[str, Any], List[FilterExpr]]] = None,
    validate_filters: bool = False,
    run_context: Optional[RunContext] = None,
    **kwargs: Any,
) -> Optional[List[Union[Dict[str, Any], str]]]:
    """Get relevant documents from knowledge base asynchronously."""
    from agno.knowledge.document import Document

    # Extract dependencies from run_context if available
    dependencies = run_context.dependencies if run_context else None

    resolved_knowledge = _get_resolved_knowledge(agent, run_context)

    if num_documents is None and resolved_knowledge is not None:
        num_documents = getattr(resolved_knowledge, "max_results", None)

    # Validate the filters against known valid filter keys
    if resolved_knowledge is not None and filters is not None:
        if validate_filters:
            valid_filters, invalid_keys = await resolved_knowledge.avalidate_filters(filters)  # type: ignore

            # Warn about invalid filter keys
            if invalid_keys:  # type: ignore
                log_warning(f"Invalid filter keys provided: {invalid_keys}. These filters will be ignored.")

                # Only use valid filters
                filters = valid_filters
                if not filters:
                    log_warning("No valid filters remain after validation. Search will proceed without filters.")

            if invalid_keys == [] and valid_filters == {}:
                log_warning("No valid filters provided. Search will proceed without filters.")
                filters = None

    if agent.knowledge_retriever is not None and callable(agent.knowledge_retriever):
        from inspect import isawaitable, signature

        try:
            sig = signature(agent.knowledge_retriever)
            knowledge_retriever_kwargs: Dict[str, Any] = {}
            if "agent" in sig.parameters:
                knowledge_retriever_kwargs = {"agent": agent}
            if "filters" in sig.parameters:
                knowledge_retriever_kwargs["filters"] = filters
            if "run_context" in sig.parameters:
                knowledge_retriever_kwargs["run_context"] = run_context
            elif "dependencies" in sig.parameters:
                # Backward compatibility: support dependencies parameter
                knowledge_retriever_kwargs["dependencies"] = dependencies
            knowledge_retriever_kwargs.update({"query": query, "num_documents": num_documents, **kwargs})
            result = agent.knowledge_retriever(**knowledge_retriever_kwargs)

            if isawaitable(result):
                result = await result

            return result
        except Exception as e:
            log_warning(f"Knowledge retriever failed: {e}")
            raise e

    # Use knowledge protocol's retrieve method
    try:
        if resolved_knowledge is None:
            return None

        # Use protocol aretrieve() or retrieve() method if available
        aretrieve_fn = getattr(resolved_knowledge, "aretrieve", None)
        retrieve_fn = getattr(resolved_knowledge, "retrieve", None)

        if not callable(aretrieve_fn) and not callable(retrieve_fn):
            log_debug("Knowledge does not implement retrieve()")
            return None

        if num_documents is None:
            num_documents = getattr(resolved_knowledge, "max_results", 10)

        log_debug(f"Retrieving from knowledge base with filters: {filters}")

        if callable(aretrieve_fn):
            relevant_docs: List[Document] = await aretrieve_fn(query=query, max_results=num_documents, filters=filters)
        elif callable(retrieve_fn):
            relevant_docs = retrieve_fn(query=query, max_results=num_documents, filters=filters)
        else:
            return None

        if not relevant_docs or len(relevant_docs) == 0:
            log_debug("No relevant documents found for query")
            return None

        return [doc.to_dict() for doc in relevant_docs]
    except Exception as e:
        log_warning(f"Error retrieving from knowledge base: {e}")
        raise e



import json
from os import getenv
from typing import Any, Dict, List, Optional, Union

import httpx
from pydantic import BaseModel, Field

from agno.tools import Toolkit
from agno.utils.log import logger


class ReplyButton(BaseModel):
    """A quick-reply button."""

    id: str = Field(..., description="Unique button identifier (e.g. 'yes', 'no').")
    title: str = Field(..., description="Button display text, max 20 characters.")


class ListRow(BaseModel):
    """A selectable row inside a list section."""

    id: str = Field(..., description="Unique row identifier.")
    title: str = Field(..., description="Row title text.")
    description: Optional[str] = Field(None, description="Optional row description.")


class ListSection(BaseModel):
    """A titled section containing selectable rows."""

    title: str = Field(..., description="Section heading.")
    rows: List[ListRow] = Field(..., description="Selectable rows in this section.")


class WhatsAppTools(Toolkit):
    def __init__(
        self,
        access_token: Optional[str] = None,
        phone_number_id: Optional[str] = None,
        version: Optional[str] = None,
        recipient_waid: Optional[str] = None,
        # Enable/disable flags
        enable_send_text_message: bool = True,
        enable_send_template_message: bool = True,
        enable_send_reply_buttons: bool = False,
        enable_send_list_message: bool = False,
        enable_send_image: bool = False,
        enable_send_document: bool = False,
        enable_send_location: bool = False,
        enable_send_reaction: bool = False,
        all: bool = False,
        **kwargs,
    ):
        self.access_token = access_token or getenv("WHATSAPP_ACCESS_TOKEN")
        if not self.access_token:
            raise ValueError("WHATSAPP_ACCESS_TOKEN is not set. Set the environment variable or pass access_token.")

        self.phone_number_id = phone_number_id or getenv("WHATSAPP_PHONE_NUMBER_ID")
        if not self.phone_number_id:
            raise ValueError(
                "WHATSAPP_PHONE_NUMBER_ID is not set. Set the environment variable or pass phone_number_id."
            )

        # Fallback recipient for standalone use outside router context
        self.default_recipient = recipient_waid or getenv("WHATSAPP_RECIPIENT_WAID")
        self.version = version or getenv("WHATSAPP_VERSION", "v22.0")
        self.base_url = "https://graph.facebook.com"

        # Register only enabled tools to keep the agent's tool list focused
        tools: List[Any] = []
        if enable_send_text_message or all:
            tools.append(self.send_text_message)
        if enable_send_template_message or all:
            tools.append(self.send_template_message)
        if enable_send_reply_buttons or all:
            tools.append(self.send_reply_buttons)
        if enable_send_list_message or all:
            tools.append(self.send_list_message)
        if enable_send_image or all:
            tools.append(self.send_image)
        if enable_send_document or all:
            tools.append(self.send_document)
        if enable_send_location or all:
            tools.append(self.send_location)
        if enable_send_reaction or all:
            tools.append(self.send_reaction)

        super().__init__(name="whatsapp", tools=tools, **kwargs)

    def _get_headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.access_token}", "Content-Type": "application/json"}

    def _get_messages_url(self) -> str:
        return f"{self.base_url}/{self.version}/{self.phone_number_id}/messages"

    def _resolve_recipient(self, recipient: Optional[str]) -> Optional[str]:
        if recipient:
            return recipient
        if self.default_recipient:
            return self.default_recipient
        return None

    def _send_message(self, data: Dict[str, Any]) -> Dict[str, Any]:
        # Raise on 4xx/5xx with parsed error body for better diagnostics
        response = httpx.post(self._get_messages_url(), headers=self._get_headers(), json=data)
        if response.status_code >= 400:
            error_body = (
                response.json()
                if response.headers.get("content-type", "").startswith("application/json")
                else {"raw": response.text}
            )
            raise httpx.HTTPStatusError(
                f"{response.status_code}: {error_body}",
                request=response.request,
                response=response,
            )
        return response.json()

    def send_text_message(
        self,
        text: str,
        recipient: Optional[str] = None,
        preview_url: bool = False,
    ) -> str:
        """Send a text message to a WhatsApp user.

        Args:
            text: The text message to send.
            recipient: Recipient's WhatsApp ID or phone number (e.g., "+1234567890"). Uses default if not provided.
            preview_url: Whether to show link previews.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            data = {
                "messaging_product": "whatsapp",
                "recipient_type": "individual",
                "to": to,
                "type": "text",
                "text": {"preview_url": preview_url, "body": text},
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending text message: {e}")
            raise

    def send_template_message(
        self,
        template_name: str,
        recipient: Optional[str] = None,
        language_code: str = "en_US",
        components: Optional[List[Dict[str, Any]]] = None,
    ) -> str:
        """Send a template message to a WhatsApp user.

        Args:
            template_name: Name of the approved template.
            recipient: Recipient's WhatsApp ID or phone number. Uses default if not provided.
            language_code: Language code for the template (e.g., "en_US").
            components: Optional list of template components (header, body, buttons).

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            template: Dict[str, Any] = {"name": template_name, "language": {"code": language_code}}
            if components:
                template["components"] = components

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "template",
                "template": template,
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending template message: {e}")
            raise

    def send_reply_buttons(
        self,
        body_text: str,
        buttons: List[ReplyButton],
        recipient: Optional[str] = None,
        header: Optional[str] = None,
        footer: Optional[str] = None,
    ) -> str:
        """Send an interactive reply button message to a WhatsApp user.

        Args:
            body_text: The message body text (max 1024 chars).
            buttons: List of ReplyButton objects, each with 'id' and 'title'. Max 3 buttons, title max 20 chars.
            recipient: Recipient phone number. Uses default if not provided.
            header: Optional header text.
            footer: Optional footer text.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            if not buttons or len(buttons) > 3:
                return json.dumps({"error": "WhatsApp requires 1-3 reply buttons"})

            action_buttons = [{"type": "reply", "reply": {"id": btn.id, "title": btn.title[:20]}} for btn in buttons]

            interactive: Dict[str, Any] = {
                "type": "button",
                "body": {"text": body_text},
                "action": {"buttons": action_buttons},
            }
            if header:
                interactive["header"] = {"type": "text", "text": header}
            if footer:
                interactive["footer"] = {"text": footer}

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "interactive",
                "interactive": interactive,
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending reply buttons: {e}")
            raise

    def send_list_message(
        self,
        body_text: str,
        button_text: str,
        sections: List[ListSection],
        recipient: Optional[str] = None,
        header: Optional[str] = None,
        footer: Optional[str] = None,
    ) -> str:
        """Send an interactive list message to a WhatsApp user.

        Args:
            body_text: The message body text.
            button_text: Text on the list button (max 20 chars).
            sections: List of ListSection objects, each with a 'title' and 'rows' list. Each row has 'id', 'title', optional 'description'.
            recipient: Recipient phone number. Uses default if not provided.
            header: Optional header text.
            footer: Optional footer text.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            if not sections or len(sections) > 10:
                return json.dumps({"error": "WhatsApp requires 1-10 sections"})

            total_rows = sum(len(s.rows) for s in sections)
            if total_rows > 10:
                return json.dumps(
                    {
                        "error": f"WhatsApp allows a maximum of 10 rows total across all sections (got {total_rows}). Reduce the number of rows."
                    }
                )

            # Build payload with truncation without mutating caller's models
            sections_payload = []
            for section in sections:
                rows = []
                for row in section.rows:
                    row_data: Dict[str, Any] = {"id": row.id, "title": row.title[:24]}
                    if row.description:
                        row_data["description"] = row.description[:72]
                    rows.append(row_data)
                sections_payload.append({"title": section.title, "rows": rows})

            interactive: Dict[str, Any] = {
                "type": "list",
                "body": {"text": body_text},
                "action": {
                    "button": button_text[:20],
                    "sections": sections_payload,
                },
            }
            if header:
                interactive["header"] = {"type": "text", "text": header}
            if footer:
                interactive["footer"] = {"text": footer}

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "interactive",
                "interactive": interactive,
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending list message: {e}")
            raise

    def send_image(
        self,
        recipient: Optional[str] = None,
        image_url: Optional[str] = None,
        media_id: Optional[str] = None,
        caption: Optional[str] = None,
    ) -> str:
        """Send an image to a WhatsApp user by URL or media ID.

        Args:
            recipient: Recipient phone number. Uses default if not provided.
            image_url: Public URL of the image. Provide either image_url or media_id.
            media_id: WhatsApp media ID from a previous upload. Provide either image_url or media_id.
            caption: Optional image caption.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            if not image_url and not media_id:
                return json.dumps({"error": "Either image_url or media_id must be provided"})

            image: Dict[str, Any] = {}
            if media_id:
                image["id"] = media_id
            else:
                image["link"] = image_url
            if caption:
                image["caption"] = caption

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "image",
                "image": image,
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending image: {e}")
            raise

    def send_document(
        self,
        recipient: Optional[str] = None,
        document_url: Optional[str] = None,
        media_id: Optional[str] = None,
        filename: Optional[str] = None,
        caption: Optional[str] = None,
    ) -> str:
        """Send a document to a WhatsApp user by URL or media ID.

        Args:
            recipient: Recipient phone number. Uses default if not provided.
            document_url: Public URL of the document. Provide either document_url or media_id.
            media_id: WhatsApp media ID from a previous upload. Provide either document_url or media_id.
            filename: Display filename for the document.
            caption: Optional document caption.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            if not document_url and not media_id:
                return json.dumps({"error": "Either document_url or media_id must be provided"})

            document: Dict[str, Any] = {}
            if media_id:
                document["id"] = media_id
            else:
                document["link"] = document_url
            if filename:
                document["filename"] = filename
            if caption:
                document["caption"] = caption

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "document",
                "document": document,
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending document: {e}")
            raise

    def send_location(
        self,
        latitude: Union[str, float],
        longitude: Union[str, float],
        name: Optional[str] = None,
        address: Optional[str] = None,
        recipient: Optional[str] = None,
    ) -> str:
        """Send a location pin to a WhatsApp user.

        Args:
            latitude: Latitude of the location (string or number).
            longitude: Longitude of the location (string or number).
            name: Optional name of the location.
            address: Optional address text.
            recipient: Recipient phone number. Uses default if not provided.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            location: Dict[str, Any] = {
                "latitude": latitude,
                "longitude": longitude,
            }
            if name:
                location["name"] = name
            if address:
                location["address"] = address

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "location",
                "location": location,
            }
            response = self._send_message(data)
            message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": message_id})
        except Exception as e:
            logger.error(f"Error sending location: {e}")
            raise

    def send_reaction(
        self,
        message_id: str,
        emoji: str,
        recipient: Optional[str] = None,
    ) -> str:
        """React to a WhatsApp message with an emoji.

        Args:
            message_id: The WhatsApp message ID of the message to react to.
            emoji: The emoji to react with (e.g., thumbs up).
            recipient: Recipient phone number. Uses default if not provided.

        Returns:
            A JSON string with message ID or error.
        """
        try:
            to = self._resolve_recipient(recipient)
            if not to:
                return json.dumps({"error": "No recipient provided and no default recipient set"})

            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "reaction",
                "reaction": {
                    "message_id": message_id,
                    "emoji": emoji,
                },
            }
            response = self._send_message(data)
            resp_message_id = response.get("messages", [{}])[0].get("id", "unknown")
            return json.dumps({"ok": True, "message_id": resp_message_id})
        except Exception as e:
            logger.error(f"Error sending reaction: {e}")
            raise



import json
from os import getenv
from typing import Any, List, Optional

from agno.tools import Toolkit
from agno.utils.log import log_debug, logger

try:
    from github import Auth, Github, GithubException

except ImportError:
    raise ImportError("`PyGithub` not installed. Please install using `pip install pygithub`")


class GithubTools(Toolkit):
    def __init__(
        self,
        access_token: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ):
        self.access_token = access_token or getenv("GITHUB_ACCESS_TOKEN")
        self.base_url = base_url

        self.g = self.authenticate()

        tools: List[Any] = [
            self.search_repositories,
            self.list_repositories,
            self.get_repository,
            self.get_pull_request,
            self.get_pull_request_changes,
            self.create_issue,
            self.create_repository,
            self.delete_repository,
            self.list_branches,
            self.get_repository_languages,
            self.get_pull_request_count,
            self.get_repository_stars,
            self.get_pull_requests,
            self.get_pull_request_comments,
            self.create_pull_request_comment,
            self.edit_pull_request_comment,
            self.get_pull_request_with_details,
            self.get_repository_with_stats,
            self.list_issues,
            self.get_issue,
            self.comment_on_issue,
            self.close_issue,
            self.reopen_issue,
            self.assign_issue,
            self.label_issue,
            self.list_issue_comments,
            self.edit_issue,
            self.create_pull_request,
            self.create_file,
            self.get_file_content,
            self.update_file,
            self.delete_file,
            self.get_directory_content,
            self.get_branch_content,
            self.create_branch,
            self.set_default_branch,
            self.search_code,
            self.search_issues_and_prs,
            self.create_review_request,
        ]

        super().__init__(name="github", tools=tools, **kwargs)

    def authenticate(self):
        """Authenticate with GitHub using the provided access token."""
        if not self.access_token:  # Fixes lint type error
            raise ValueError("GitHub access token is required")

        auth = Auth.Token(self.access_token)
        if self.base_url:
            log_debug(f"Authenticating with GitHub Enterprise at {self.base_url}")
            return Github(base_url=self.base_url, auth=auth)
        else:
            log_debug("Authenticating with public GitHub")
            return Github(auth=auth)

    def search_repositories(
        self,
        query: str,
        sort: str = "stars",
        order: str = "desc",
        page: int = 1,
        per_page: int = 30,
    ) -> str:
        """Search for repositories on GitHub.

        Note: GitHub's Search API has a maximum limit of 1000 results per query.

        Args:
            query (str): The search query keywords.
            sort (str, optional): The field to sort results by. Can be 'stars', 'forks', or 'updated'. Defaults to 'stars'.
            order (str, optional): The order of results. Can be 'asc' or 'desc'. Defaults to 'desc'.
            page (int, optional): Page number of results to return, counting from 1. Defaults to 1.
            per_page (int, optional): Number of results per page. Max 100. Defaults to 30.

        Returns:
            A JSON-formatted string containing a list of repositories matching the search query.
        """
        log_debug(f"Searching repositories with query: '{query}', page: {page}, per_page: {per_page}")
        try:
            # Ensure per_page doesn't exceed GitHub's max of 100
            per_page = min(per_page, 100)

            repositories = self.g.search_repositories(query=query, sort=sort, order=order)

            # Get the specified page of results
            repo_list = []
            for repo in repositories.get_page(page - 1):
                repo_info = {
                    "full_name": repo.full_name,
                    "description": repo.description,
                    "url": repo.html_url,
                    "stars": repo.stargazers_count,
                    "forks": repo.forks_count,
                    "language": repo.language,
                }
                repo_list.append(repo_info)

                if len(repo_list) >= per_page:
                    break

            return json.dumps(repo_list, indent=2)

        except GithubException as e:
            logger.error(f"Error searching repositories: {e}")
            return json.dumps({"error": str(e)})

    def list_repositories(self) -> str:
        """List all repositories for the authenticated user.

        Returns:
            A JSON-formatted string containing a list of repository names.
        """
        log_debug("Listing repositories")
        try:
            repos = self.g.get_user().get_repos()
            repo_names = [repo.full_name for repo in repos]
            return json.dumps(repo_names, indent=2)
        except GithubException as e:
            logger.error(f"Error listing repositories: {e}")
            return json.dumps({"error": str(e)})

    def create_repository(
        self,
        name: str,
        private: bool = False,
        description: Optional[str] = None,
        auto_init: bool = False,
        organization: Optional[str] = None,
    ) -> str:
        """Create a new repository on GitHub.

        Args:
            name (str): The name of the repository.
            private (bool, optional): Whether the repository is private. Defaults to False.
            description (str, optional): A short description of the repository.
            auto_init (bool, optional): Whether to initialize the repository with a README. Defaults to False.
            organization (str, optional): Name of organization to create repo in. If None, creates in user account.

        Returns:
            A JSON-formatted string containing the created repository details.
        """
        log_debug(f"Creating repository: {name}")
        try:
            description = description if description is not None else ""

            if organization:
                log_debug(f"Creating in organization: {organization}")
                org = self.g.get_organization(organization)
                repo = org.create_repo(
                    name=name,
                    private=private,
                    description=description,
                    auto_init=auto_init,
                )
            else:
                repo = self.g.get_user().create_repo(
                    name=name,
                    private=private,
                    description=description,
                    auto_init=auto_init,
                )

            repo_info = {
                "name": repo.full_name,
                "url": repo.html_url,
                "private": repo.private,
                "description": repo.description,
            }
            return json.dumps(repo_info, indent=2)
        except GithubException as e:
            logger.error(f"Error creating repository: {e}")
            return json.dumps({"error": str(e)})

    def get_repository(self, repo_name: str) -> str:
        """Get details of a specific repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').

        Returns:
            A JSON-formatted string containing repository details.
        """
        log_debug(f"Getting repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            repo_info = {
                "name": repo.full_name,
                "description": repo.description,
                "url": repo.html_url,
                "stars": repo.stargazers_count,
                "forks": repo.forks_count,
                "open_issues": repo.open_issues_count,
                "language": repo.language,
                "license": repo.license.name if repo.license else None,
                "default_branch": repo.default_branch,
            }
            return json.dumps(repo_info, indent=2)
        except GithubException as e:
            logger.error(f"Error getting repository: {e}")
            return json.dumps({"error": str(e)})

    def get_repository_languages(self, repo_name: str) -> str:
        """Get the languages used in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').

        Returns:
            A JSON-formatted string containing the list of languages.
        """
        log_debug(f"Getting languages for repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            languages = repo.get_languages()
            return json.dumps(languages, indent=2)
        except GithubException as e:
            logger.error(f"Error getting repository languages: {e}")
            return json.dumps({"error": str(e)})

    def get_pull_request_count(
        self,
        repo_name: str,
        state: str = "all",
        author: Optional[str] = None,
        base: Optional[str] = None,
        head: Optional[str] = None,
    ) -> str:
        """Get the count of pull requests for a repository based on query parameters.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            state (str, optional): The state of the PRs to count ('open', 'closed', 'all'). Defaults to 'all'.
            author (str, optional): Filter PRs by author username.
            base (str, optional): Filter PRs by base branch name.
            head (str, optional): Filter PRs by head branch name.

        Returns:
            A JSON-formatted string containing the count of pull requests.
        """
        log_debug(f"Counting pull requests for repository: {repo_name} with state: {state}")
        try:
            repo = self.g.get_repo(repo_name)
            pulls = repo.get_pulls(state=state, base=base, head=head)

            # If author is specified, filter the results
            if author:
                # If we need to filter by author and state, make sure both conditions are met
                if state != "all":
                    count = sum(1 for pr in pulls if pr.user.login == author and pr.state == state)
                else:
                    count = sum(1 for pr in pulls if pr.user.login == author)
            else:
                count = pulls.totalCount

            return json.dumps({"count": count}, indent=2)
        except GithubException as e:
            logger.error(f"Error counting pull requests: {e}")
            return json.dumps({"error": str(e)})

    def get_pull_request(self, repo_name: str, pr_number: int) -> str:
        """Get details of a specific pull request.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            pr_number (int): The number of the pull request.


        Returns:
            A JSON-formatted string containing pull request details.
        """
        log_debug(f"Getting pull request #{pr_number} for repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.get_pull(pr_number)
            pr_info = {
                "number": pr.number,
                "title": pr.title,
                "user": pr.user.login,
                "body": pr.body,
                "created_at": pr.created_at.isoformat(),
                "updated_at": pr.updated_at.isoformat(),
                "state": pr.state,
                "merged": pr.is_merged(),
                "mergeable": pr.mergeable,
                "url": pr.html_url,
            }
            return json.dumps(pr_info, indent=2)
        except GithubException as e:
            logger.error(f"Error getting pull request: {e}")
            return json.dumps({"error": str(e)})

    def get_pull_request_changes(self, repo_name: str, pr_number: int) -> str:
        """Get the changes (files modified) in a pull request.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            pr_number (int): The number of the pull request.

        Returns:
            A JSON-formatted string containing the list of changed files.
        """
        log_debug(f"Getting changes for pull request #{pr_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.get_pull(pr_number)
            files = pr.get_files()
            changes = []
            for file in files:
                file_info = {
                    "filename": file.filename,
                    "status": file.status,
                    "additions": file.additions,
                    "deletions": file.deletions,
                    "changes": file.changes,
                    "raw_url": file.raw_url,
                    "blob_url": file.blob_url,
                    "patch": file.patch,
                }
                changes.append(file_info)
            return json.dumps(changes, indent=2)
        except GithubException as e:
            logger.error(f"Error getting pull request changes: {e}")
            return json.dumps({"error": str(e)})

    def create_issue(self, repo_name: str, title: str, body: Optional[str] = None) -> str:
        """Create an issue in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            title (str): The title of the issue.
            body (str, optional): The body content of the issue.

        Returns:
            A JSON-formatted string containing the created issue details.
        """
        log_debug(f"Creating issue in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.create_issue(title=title, body=body)  # type: ignore
            issue_info = {
                "id": issue.id,
                "number": issue.number,
                "title": issue.title,
                "body": issue.body,
                "url": issue.html_url,
                "state": issue.state,
                "created_at": issue.created_at.isoformat(),
                "user": issue.user.login,
            }
            return json.dumps(issue_info, indent=2)
        except GithubException as e:
            logger.error(f"Error creating issue: {e}")
            return json.dumps({"error": str(e)})

    def list_issues(self, repo_name: str, state: str = "open", page: int = 1, per_page: int = 20) -> str:
        """List issues for a repository with pagination.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            state (str, optional): The state of issues to list ('open', 'closed', 'all'). Defaults to 'open'.
            page (int, optional): Page number of results to return, counting from 1. Defaults to 1.
            per_page (int, optional): Number of results per page. Defaults to 20.
        Returns:
            A JSON-formatted string containing a list of issues with pagination metadata.
        """
        log_debug(f"Listing issues for repository: {repo_name} with state: {state}, page: {page}, per_page: {per_page}")
        try:
            repo = self.g.get_repo(repo_name)

            issues = repo.get_issues(state=state)

            # Filter out pull requests after fetching issues
            total_issues = 0
            all_issues = []
            for issue in issues:
                if not issue.pull_request:
                    all_issues.append(issue)
                    total_issues += 1

            # Calculate pagination metadata
            total_pages = (total_issues + per_page - 1) // per_page

            # Validate page number
            if page < 1:
                page = 1
            elif page > total_pages and total_pages > 0:
                page = total_pages

            # Get the specified page of results
            issue_list = []
            page_start = (page - 1) * per_page
            page_end = page_start + per_page

            for i in range(page_start, min(page_end, total_issues)):
                if i < len(all_issues):
                    issue = all_issues[i]
                    issue_info = {
                        "number": issue.number,
                        "title": issue.title,
                        "user": issue.user.login,
                        "created_at": issue.created_at.isoformat(),
                        "state": issue.state,
                        "url": issue.html_url,
                    }
                    issue_list.append(issue_info)

            meta = {"current_page": page, "per_page": per_page, "total_items": total_issues, "total_pages": total_pages}

            response = {"data": issue_list, "meta": meta}

            return json.dumps(response, indent=2)
        except GithubException as e:
            logger.error(f"Error listing issues: {e}")
            return json.dumps({"error": str(e)})

    def get_issue(self, repo_name: str, issue_number: int) -> str:
        """Get details of a specific issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.

        Returns:
            A JSON-formatted string containing issue details.
        """
        log_debug(f"Getting issue #{issue_number} for repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            issue_info = {
                "number": issue.number,
                "title": issue.title,
                "body": issue.body,
                "user": issue.user.login,
                "state": issue.state,
                "created_at": issue.created_at.isoformat(),
                "updated_at": issue.updated_at.isoformat(),
                "url": issue.html_url,
                "assignees": [assignee.login for assignee in issue.assignees],
                "labels": [label.name for label in issue.labels],
            }
            return json.dumps(issue_info, indent=2)
        except GithubException as e:
            logger.error(f"Error getting issue: {e}")
            return json.dumps({"error": str(e)})

    def comment_on_issue(self, repo_name: str, issue_number: int, comment_body: str) -> str:
        """Add a comment to an issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.
            comment_body (str): The content of the comment.

        Returns:
            A JSON-formatted string containing the comment details.
        """
        log_debug(f"Adding comment to issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            comment = issue.create_comment(body=comment_body)
            comment_info = {
                "id": comment.id,
                "body": comment.body,
                "user": comment.user.login,
                "created_at": comment.created_at.isoformat(),
                "url": comment.html_url,
            }
            return json.dumps(comment_info, indent=2)
        except GithubException as e:
            logger.error(f"Error commenting on issue: {e}")
            return json.dumps({"error": str(e)})

    def close_issue(self, repo_name: str, issue_number: int) -> str:
        """Close an issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.

        Returns:
            A JSON-formatted string confirming the issue is closed.
        """
        log_debug(f"Closing issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            issue.edit(state="closed")
            return json.dumps({"message": f"Issue #{issue_number} closed."}, indent=2)
        except GithubException as e:
            logger.error(f"Error closing issue: {e}")
            return json.dumps({"error": str(e)})

    def reopen_issue(self, repo_name: str, issue_number: int) -> str:
        """Reopen a closed issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.

        Returns:
            A JSON-formatted string confirming the issue is reopened.
        """
        log_debug(f"Reopening issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            issue.edit(state="open")
            return json.dumps({"message": f"Issue #{issue_number} reopened."}, indent=2)
        except GithubException as e:
            logger.error(f"Error reopening issue: {e}")
            return json.dumps({"error": str(e)})

    def assign_issue(self, repo_name: str, issue_number: int, assignees: List[str]) -> str:
        """Assign users to an issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.
            assignees (List[str]): A list of usernames to assign.

        Returns:
            A JSON-formatted string confirming the assignees.
        """
        log_debug(f"Assigning users to issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            issue.edit(assignees=assignees)
            return json.dumps({"message": f"Issue #{issue_number} assigned to {assignees}."}, indent=2)
        except GithubException as e:
            logger.error(f"Error assigning issue: {e}")
            return json.dumps({"error": str(e)})

    def label_issue(self, repo_name: str, issue_number: int, labels: List[str]) -> str:
        """Add labels to an issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.
            labels (List[str]): A list of label names to add.

        Returns:
            A JSON-formatted string confirming the labels.
        """
        log_debug(f"Labeling issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            issue.edit(labels=labels)
            return json.dumps(
                {"message": f"Labels {labels} added to issue #{issue_number}."},
                indent=2,
            )
        except GithubException as e:
            logger.error(f"Error labeling issue: {e}")
            return json.dumps({"error": str(e)})

    def list_issue_comments(self, repo_name: str, issue_number: int) -> str:
        """List comments on an issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.

        Returns:
            A JSON-formatted string containing a list of comments.
        """
        log_debug(f"Listing comments for issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            comments = issue.get_comments()
            comment_list = []
            for comment in comments:
                comment_info = {
                    "id": comment.id,
                    "user": comment.user.login,
                    "body": comment.body,
                    "created_at": comment.created_at.isoformat(),
                    "url": comment.html_url,
                }
                comment_list.append(comment_info)
            return json.dumps(comment_list, indent=2)
        except GithubException as e:
            logger.error(f"Error listing issue comments: {e}")
            return json.dumps({"error": str(e)})

    def edit_issue(
        self,
        repo_name: str,
        issue_number: int,
        title: Optional[str] = None,
        body: Optional[str] = None,
    ) -> str:
        """Edit the title or body of an issue.

        Args:
            repo_name (str): The full name of the repository.
            issue_number (int): The number of the issue.
            title (str, optional): The new title for the issue.
            body (str, optional): The new body content for the issue.

        Returns:
            A JSON-formatted string confirming the issue has been updated.
        """
        log_debug(f"Editing issue #{issue_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            issue = repo.get_issue(number=issue_number)
            issue.edit(title=title, body=body)  # type: ignore
            return json.dumps({"message": f"Issue #{issue_number} updated."}, indent=2)
        except GithubException as e:
            logger.error(f"Error editing issue: {e}")
            return json.dumps({"error": str(e)})

    def delete_repository(self, repo_name: str) -> str:
        """Delete a repository (requires admin permissions).

            Args:
                repo_name (str): The full name of the repository to delete (e.g., 'owner/repo').

        Returns:
            A JSON-formatted string with success message or error.
        """
        log_debug(f"Deleting repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            repo.delete()
            return json.dumps({"message": f"Repository {repo_name} deleted successfully"}, indent=2)
        except GithubException as e:
            logger.error(f"Error deleting repository: {e}")
            return json.dumps({"error": str(e)})

    def list_branches(self, repo_name: str) -> str:
        """List all branches in a repository.

        Args:
            repo_name (str): Full repository name (e.g., 'owner/repo').

        Returns:
            JSON list of branch names.
        """
        try:
            repo = self.g.get_repo(repo_name)
            branches = [branch.name for branch in repo.get_branches()]
            return json.dumps(branches, indent=2)
        except GithubException as e:
            logger.error(f"Error listing branches: {e}")
            return json.dumps({"error": str(e)})

    def get_repository_stars(self, repo_name: str) -> str:
        """Get the number of stars for a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').

        Returns:
            A JSON-formatted string containing the star count.
        """
        log_debug(f"Getting star count for repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            return json.dumps({"stars": repo.stargazers_count}, indent=2)
        except GithubException as e:
            logger.error(f"Error getting repository stars: {e}")
            return json.dumps({"error": str(e)})

    def get_pull_requests(
        self,
        repo_name: str,
        state: str = "open",
        sort: str = "created",
        direction: str = "desc",
        limit: int = 50,
    ) -> str:
        """Get pull requests matching query parameters.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            state (str, optional): State of the PRs to retrieve. Can be 'open', 'closed', or 'all'. Defaults to 'open'.
            sort (str, optional): What to sort results by. Can be 'created', 'updated', 'popularity', 'long-running'. Defaults to 'created'.
            direction (str, optional): The direction of the sort. Can be 'asc' or 'desc'. Defaults to 'desc'.
            limit (int, optional): The maximum number of pull requests to return. Defaults to 20.

        Returns:
            A JSON-formatted string containing a list of pull requests.
        """
        try:
            repo = self.g.get_repo(repo_name)
            pulls = repo.get_pulls(state=state, sort=sort, direction=direction)

            pr_list = []
            for pr in pulls[:limit]:
                pr_info = {
                    "number": pr.number,
                    "title": pr.title,
                    "user": pr.user.login,
                    "created_at": pr.created_at.isoformat(),
                    "updated_at": pr.updated_at.isoformat(),
                    "state": pr.state,
                    "url": pr.html_url,
                }
                pr_list.append(pr_info)

            return json.dumps(pr_list, indent=2)
        except GithubException as e:
            logger.error(f"Error getting pull requests by query: {e}")
            return json.dumps({"error": str(e)})

    def get_pull_request_comments(self, repo_name: str, pr_number: int, include_issue_comments: bool = True) -> str:
        """Get all comments on a pull request.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            pr_number (int): The number of the pull request.
            include_issue_comments (bool, optional): Whether to include general PR comments. Defaults to True.

        Returns:
            A JSON-formatted string containing a list of pull request comments.
        """
        log_debug(f"Getting comments for pull request #{pr_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.get_pull(pr_number)

            comment_list = []

            # Get review comments (comments on specific lines of code)
            review_comments = pr.get_comments()
            for comment in review_comments:
                comment_info = {
                    "id": comment.id,
                    "body": comment.body,
                    "user": comment.user.login,
                    "created_at": comment.created_at.isoformat(),
                    "updated_at": comment.updated_at.isoformat(),
                    "path": comment.path,
                    "position": comment.position,
                    "commit_id": comment.commit_id,
                    "url": comment.html_url,
                    "type": "review_comment",
                }
                comment_list.append(comment_info)

            # Get general issue comments if requested
            if include_issue_comments:
                issue_comments = pr.get_issue_comments()
                for comment in issue_comments:
                    comment_info = {
                        "id": comment.id,
                        "body": comment.body,
                        "user": comment.user.login,
                        "created_at": comment.created_at.isoformat(),
                        "updated_at": comment.updated_at.isoformat(),
                        "url": comment.html_url,
                        "type": "issue_comment",
                    }
                    comment_list.append(comment_info)

            # Sort all comments by creation date
            comment_list.sort(key=lambda x: x["created_at"], reverse=True)

            return json.dumps(comment_list, indent=2)
        except GithubException as e:
            logger.error(f"Error getting pull request comments: {e}")
            return json.dumps({"error": str(e)})

    def create_pull_request_comment(
        self,
        repo_name: str,
        pr_number: int,
        body: str,
        commit_id: str,
        path: str,
        position: int,
    ) -> str:
        """Create a comment on a specific line of a specific file in a pull request.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            pr_number (int): The number of the pull request.
            body (str): The text of the comment.
            commit_id (str): The SHA of the commit to comment on.
            path (str): The relative path to the file to comment on.
            position (int): The line index in the diff to comment on.

        Returns:
            A JSON-formatted string containing the created comment details.
        """
        log_debug(f"Creating comment on pull request #{pr_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.get_pull(pr_number)
            commit = repo.get_commit(commit_id)
            comment = pr.create_comment(body, commit, path, position)

            comment_info = {
                "id": comment.id,
                "body": comment.body,
                "user": comment.user.login,
                "created_at": comment.created_at.isoformat(),
                "path": comment.path,
                "position": comment.position,
                "commit_id": comment.commit_id,
                "url": comment.html_url,
            }

            return json.dumps(comment_info, indent=2)
        except GithubException as e:
            logger.error(f"Error creating pull request comment: {e}")
            return json.dumps({"error": str(e)})

    def edit_pull_request_comment(self, repo_name: str, comment_id: int, body: str) -> str:
        """Edit an existing pull request comment.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            comment_id (int): The id of the comment to edit.
            body (str): The new text of the comment.

        Returns:
            A JSON-formatted string containing the updated comment details.
        """
        log_debug(f"Editing comment #{comment_id} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            comments = repo.get_pulls_comments()
            comment = None
            for comment in comments:
                if comment.id == comment_id:
                    comment.edit(body)

            if not comment:
                return f"Could not find comment #{comment_id} in repository: {repo_name}"

            comment_info = {
                "id": comment.id,
                "body": comment.body,
                "user": comment.user.login,
                "updated_at": comment.updated_at.isoformat(),
                "path": comment.path,
                "position": comment.position,
                "commit_id": comment.commit_id,
                "url": comment.html_url,
            }

            return json.dumps(comment_info, indent=2)
        except GithubException as e:
            logger.error(f"Error editing pull request comment: {e}")
            return json.dumps({"error": str(e)})

    def get_pull_request_with_details(self, repo_name: str, pr_number: int) -> str:
        """Get comprehensive details of a pull request including comments, labels, and metadata.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            pr_number (int): The number of the pull request.

        Returns:
            A JSON-formatted string containing detailed pull request information.
        """
        log_debug(f"Getting comprehensive details for PR #{pr_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.get_pull(pr_number)

            # Get review comments
            review_comments = []
            for comment in pr.get_comments():
                review_comments.append(
                    {
                        "id": comment.id,
                        "body": comment.body,
                        "user": comment.user.login,
                        "created_at": comment.created_at.isoformat(),
                        "path": comment.path,
                        "position": comment.position,
                        "commit_id": comment.commit_id,
                        "url": comment.html_url,
                        "type": "review_comment",
                    }
                )

            # Get issue comments
            issue_comments = []
            for comment in pr.get_issue_comments():
                issue_comments.append(
                    {
                        "id": comment.id,
                        "body": comment.body,
                        "user": comment.user.login,
                        "created_at": comment.created_at.isoformat(),
                        "url": comment.html_url,
                        "type": "issue_comment",
                    }
                )

            # Get commit data
            commits = []
            for commit in pr.get_commits():
                commit_info = {
                    "sha": commit.sha,
                    "message": commit.commit.message,
                    "author": (commit.commit.author.name if commit.commit.author else "Unknown"),
                    "date": (commit.commit.author.date.isoformat() if commit.commit.author else None),
                    "url": commit.html_url,
                }
                commits.append(commit_info)

            # Get files changed
            files_changed = []
            for file in pr.get_files():
                file_info = {
                    "filename": file.filename,
                    "status": file.status,
                    "additions": file.additions,
                    "deletions": file.deletions,
                    "changes": file.changes,
                    "patch": file.patch,
                }
                files_changed.append(file_info)

            # Combine all comments and sort by creation date
            all_comments = review_comments + issue_comments
            all_comments.sort(key=lambda x: x["created_at"], reverse=True)

            # Get basic PR info
            pr_info = {
                "number": pr.number,
                "title": pr.title,
                "user": pr.user.login,
                "state": pr.state,
                "created_at": pr.created_at.isoformat(),
                "updated_at": pr.updated_at.isoformat(),
                "html_url": pr.html_url,
                "body": pr.body,
                "base": pr.base.ref,
                "head": pr.head.ref,
                "merged": pr.is_merged(),
                "mergeable": pr.mergeable,
                "additions": pr.additions,
                "deletions": pr.deletions,
                "changed_files": pr.changed_files,
                "labels": [label.name for label in pr.labels],
                "comments_count": {
                    "review_comments": len(review_comments),
                    "issue_comments": len(issue_comments),
                    "total": len(all_comments),
                },
                "comments": all_comments,
                "commits": commits,
                "files_changed": files_changed,
            }

            return json.dumps(pr_info, indent=2)
        except GithubException as e:
            logger.error(f"Error getting pull request details: {e}")
            return json.dumps({"error": str(e)})

    def get_repository_with_stats(self, repo_name: str) -> str:
        """Get comprehensive repository information including statistics.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').

        Returns:
            A JSON-formatted string containing detailed repository information and statistics.
        """
        log_debug(f"Getting detailed info for repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Helper function to safely convert values to primitive types
            def safe_value(val):
                if hasattr(val, "isoformat"):
                    return val.isoformat()
                elif isinstance(val, (int, float, bool, str)) or val is None:
                    return val
                else:
                    return str(val)

            # Get basic repo info
            repo_info = {
                "id": int(repo.id),
                "name": str(repo.name),
                "full_name": str(repo.full_name),
                "owner": str(repo.owner.login),
                "description": str(repo.description) if repo.description else None,
                "html_url": str(repo.html_url),
                "homepage": str(repo.homepage) if repo.homepage else None,
                "language": str(repo.language) if repo.language else None,
                "created_at": safe_value(repo.created_at),
                "updated_at": safe_value(repo.updated_at),
                "pushed_at": safe_value(repo.pushed_at),
                "size": int(repo.size),
                "stargazers_count": int(repo.stargazers_count),
                "watchers_count": int(repo.watchers_count),
                "forks_count": int(repo.forks_count),
                "open_issues_count": int(repo.open_issues_count),
                "default_branch": str(repo.default_branch),
                "topics": [str(topic) for topic in repo.get_topics()],
                "license": (str(repo.license.name) if repo.license and hasattr(repo.license, "name") else None),
                "private": bool(repo.private),
                "archived": bool(repo.archived),
            }

            # Get languages
            repo_info["languages"] = {str(lang): int(count) for lang, count in repo.get_languages().items()}

            # Calculate actual open issues (GitHub's count includes PRs)
            try:
                open_issues_count = 0
                for issue in repo.get_issues(state="open"):
                    if not issue.pull_request:
                        open_issues_count += 1
                repo_info["actual_open_issues"] = open_issues_count
            except Exception as e:
                log_debug(f"Error getting actual open issues: {e}")
                repo_info["actual_open_issues"] = None

            # Get open pull requests count
            try:
                open_prs = repo.get_pulls(state="open")
                repo_info["open_pr_count"] = int(open_prs.totalCount)
            except Exception as e:
                log_debug(f"Error getting open PRs count: {e}")
                repo_info["open_pr_count"] = None

            # Get recent open PRs
            try:
                open_prs_list = []
                open_prs = repo.get_pulls(state="open")

                # Use a simple for loop approach instead of trying to slice first
                count = 0
                for pr in open_prs:
                    if count >= 10:
                        break
                    try:
                        # Ensure all fields are primitives, not Mock objects
                        pr_data = {
                            "number": int(pr.number),
                            "title": str(pr.title),
                            "user": str(pr.user.login),
                            "created_at": safe_value(pr.created_at),
                            "updated_at": safe_value(pr.updated_at),
                            "url": str(pr.html_url),
                            "base": str(pr.base.ref),
                            "head": str(pr.head.ref),
                            "comment_count": int(pr.comments),
                        }
                        open_prs_list.append(pr_data)
                        count += 1
                    except Exception as e:
                        log_debug(f"Error processing individual PR: {e}")

                repo_info["recent_open_prs"] = open_prs_list
            except Exception as e:
                log_debug(f"Error getting recent open PRs: {e}")
                repo_info["recent_open_prs"] = []

            # Calculate PR metrics
            try:
                # Get a sample of PRs for statistics
                all_prs_list = []
                all_prs = repo.get_pulls(state="all", sort="created", direction="desc")

                pr_count = 0
                for pr in all_prs:
                    if pr_count >= 100:  # Limit to 100 PRs
                        break
                    all_prs_list.append(pr)
                    pr_count += 1

                # Calculate basic metrics
                merged_prs = []
                for pr in all_prs_list:
                    is_merged = pr.is_merged()
                    if is_merged:
                        merged_prs.append(pr)

                # Compute merge time for merged PRs (in hours)
                merge_times = []
                for pr in merged_prs:
                    if pr.merged_at and pr.created_at:
                        merge_time = (pr.merged_at - pr.created_at).total_seconds() / 3600
                        merge_times.append(merge_time)

                pr_metrics = {
                    "total_prs": len(all_prs_list),
                    "merged_prs": len(merged_prs),
                    "acceptance_rate": ((len(merged_prs) / len(all_prs_list) * 100) if len(all_prs_list) > 0 else 0),
                    "avg_time_to_merge": (sum(merge_times) / len(merge_times) if merge_times else None),
                }
                repo_info["pr_metrics"] = pr_metrics
            except Exception as e:
                log_debug(f"Error calculating PR metrics: {e}")
                repo_info["pr_metrics"] = None

            # Get contributors
            try:
                contributors: list[dict] = []
                for contributor in repo.get_contributors():
                    if len(contributors) >= 20:  # Limit to top 20
                        break
                    contributors.append(
                        {
                            "login": str(contributor.login),
                            "contributions": int(contributor.contributions),
                            "url": str(contributor.html_url),
                        }
                    )
                repo_info["contributors"] = contributors
            except Exception as e:
                log_debug(f"Error getting contributors: {e}")
                repo_info["contributors"] = []

            return json.dumps(repo_info, indent=2)
        except GithubException as e:
            logger.error(f"Error getting repository stats: {e}")
            return json.dumps({"error": str(e)})

    def create_pull_request(
        self,
        repo_name: str,
        title: str,
        body: str,
        head: str,
        base: str,
        draft: bool = False,
        maintainer_can_modify: bool = True,
    ) -> str:
        """Create a new pull request in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            title (str): The title of the pull request.
            body (str): The body text of the pull request.
            head (str): The name of the branch where your changes are implemented.
            base (str): The name of the branch you want the changes pulled into.
            draft (bool, optional): Whether the pull request is a draft. Defaults to False.
            maintainer_can_modify (bool, optional): Whether maintainers can modify the PR. Defaults to True.

        Returns:
            A JSON-formatted string containing the created pull request details.
        """
        log_debug(f"Creating pull request in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.create_pull(
                title=title,
                body=body,
                head=head,
                base=base,
                draft=draft,
                maintainer_can_modify=maintainer_can_modify,
            )

            pr_info = {
                "number": pr.number,
                "title": pr.title,
                "body": pr.body,
                "user": pr.user.login,
                "state": pr.state,
                "created_at": pr.created_at.isoformat(),
                "html_url": pr.html_url,
                "base": pr.base.ref,
                "head": pr.head.ref,
                "mergeable": pr.mergeable,
            }

            return json.dumps(pr_info, indent=2)
        except GithubException as e:
            logger.error(f"Error creating pull request: {e}")
            return json.dumps({"error": str(e)})

    def create_review_request(
        self,
        repo_name: str,
        pr_number: int,
        reviewers: List[str],
        team_reviewers: Optional[List[str]] = None,
    ) -> str:
        """Create a review request for a pull request.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            pr_number (int): The number of the pull request.
            reviewers (List[str]): List of user logins that will be requested to review.
            team_reviewers (List[str], optional): List of team slugs that will be requested to review. Defaults to None.

        Returns:
            A JSON-formatted string with the success message or error.
        """
        log_debug(f"Creating review request for PR #{pr_number} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)
            pr = repo.get_pull(pr_number)
            pr.create_review_request(reviewers=reviewers, team_reviewers=team_reviewers or [])

            return json.dumps(
                {
                    "message": f"Review request created for PR #{pr_number}",
                    "requested_reviewers": reviewers,
                    "requested_team_reviewers": team_reviewers or [],
                },
                indent=2,
            )
        except GithubException as e:
            logger.error(f"Error creating review request: {e}")
            return json.dumps({"error": str(e)})

    def create_file(
        self,
        repo_name: str,
        path: str,
        content: str,
        message: str,
        branch: Optional[str] = None,
    ) -> str:
        """Create a new file in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            path (str): The path to the file in the repository.
            content (str): The content of the file.
            message (str): The commit message.
            branch (str, optional): The branch to commit to. Defaults to repository's default branch.

        Returns:
            A JSON-formatted string containing the file creation result.
        """
        log_debug(f"Creating file {path} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Convert string content to bytes
            content_bytes = content.encode("utf-8")

            # Create the file
            result = repo.create_file(path=path, message=message, content=content_bytes, branch=branch)

            # Extract relevant information
            file_info = {
                "path": result["content"].path,  # type: ignore
                "sha": result["content"].sha,
                "url": result["content"].html_url,
                "commit": {
                    "sha": result["commit"].sha,
                    "message": result["commit"].commit.message
                    if result["commit"].commit
                    else result["commit"]._rawData["message"],
                    "url": result["commit"].html_url,
                },
            }

            return json.dumps(file_info, indent=2)
        except (GithubException, AssertionError) as e:
            logger.error(f"Error creating file: {e}")
            return json.dumps({"error": str(e)})

    def get_file_content(self, repo_name: str, path: str, ref: Optional[str] = None) -> str:
        """Get the content of a file in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            path (str): The path to the file in the repository.
            ref (str, optional): The name of the commit/branch/tag. Defaults to the repository's default branch.

        Returns:
            A JSON-formatted string containing the file content and metadata.
        """
        log_debug(f"Getting content of file {path} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Conditionally call get_contents based on ref
            if ref is not None:
                file_content = repo.get_contents(path, ref=ref)
            else:
                file_content = repo.get_contents(path)

            # If it's a list (directory), raise an error
            if isinstance(file_content, list):
                return json.dumps({"error": f"{path} is a directory, not a file"})

            # Decode content
            try:
                decoded_content = file_content.decoded_content.decode("utf-8")
            except UnicodeDecodeError:
                decoded_content = "Binary file (content not displayed)"
            except Exception as e:
                log_debug(f"Error decoding file content: {e}")
                decoded_content = "Binary file (content not displayed)"

            # Make sure we don't try to display binary content
            if isinstance(decoded_content, str) and (
                "\x00" in decoded_content or sum(1 for c in decoded_content[:1000] if not (32 <= ord(c) <= 126)) > 200
            ):
                decoded_content = "Binary file (content not displayed)"

            # Create response
            content_info = {
                "name": file_content.name,
                "path": file_content.path,
                "sha": file_content.sha,
                "size": file_content.size,
                "type": file_content.type,
                "url": file_content.html_url,
                "content": decoded_content,
            }

            return json.dumps(content_info, indent=2)
        except GithubException as e:
            logger.error(f"Error getting file content: {e}")
            return json.dumps({"error": str(e)})

    def update_file(
        self,
        repo_name: str,
        path: str,
        content: str,
        message: str,
        sha: str,
        branch: Optional[str] = None,
    ) -> str:
        """Update an existing file in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            path (str): The path to the file in the repository.
            content (str): The new content of the file.
            message (str): The commit message.
            sha (str): The blob SHA of the file being replaced.
            branch (str, optional): The branch to commit to. Defaults to repository's default branch.

        Returns:
            A JSON-formatted string containing the file update result.
        """
        log_debug(f"Updating file {path} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Convert string content to bytes
            content_bytes = content.encode("utf-8")

            # Update the file
            result = repo.update_file(
                path=path,
                message=message,
                content=content_bytes,
                sha=sha,
                branch=branch,
            )

            # Extract relevant information
            file_info = {
                "path": result["content"].path,
                "sha": result["content"].sha,
                "url": result["content"].html_url,
                "commit": {
                    "sha": result["commit"].sha,
                    "message": result["commit"].commit.message,
                    "url": result["commit"].html_url,
                },
            }

            return json.dumps(file_info, indent=2)
        except GithubException as e:
            logger.error(f"Error updating file: {e}")
            return json.dumps({"error": str(e)})

    def delete_file(
        self,
        repo_name: str,
        path: str,
        message: str,
        sha: str,
        branch: Optional[str] = None,
    ) -> str:
        """Delete a file from a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            path (str): The path to the file in the repository.
            message (str): The commit message.
            sha (str): The blob SHA of the file being deleted.
            branch (str, optional): The branch to commit to. Defaults to repository's default branch.

        Returns:
            A JSON-formatted string containing the file deletion result.
        """
        log_debug(f"Deleting file {path} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Delete the file
            result = repo.delete_file(path=path, message=message, sha=sha, branch=branch)

            # Extract relevant information
            commit_info = {
                "message": f"File {path} deleted successfully",
                "commit": {
                    "sha": result["commit"].sha,
                    "message": result["commit"].commit.message,
                    "url": result["commit"].html_url,
                },
            }

            return json.dumps(commit_info, indent=2)
        except GithubException as e:
            logger.error(f"Error deleting file: {e}")
            return json.dumps({"error": str(e)})

    def get_directory_content(self, repo_name: str, path: str, ref: Optional[str] = None) -> str:
        """Get the contents of a directory in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            path (str): The path to the directory in the repository. Use empty string for root.
            ref (str, optional): The name of the commit/branch/tag. Defaults to repository's default branch.

        Returns:
            A JSON-formatted string containing a list of directory contents.
        """
        log_debug(f"Getting contents of directory {path} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Conditionally call get_contents based on ref
            if ref is not None:
                contents = repo.get_contents(path, ref=ref)
            else:
                contents = repo.get_contents(path)

            # If it's not a list, it's a file not a directory
            if not isinstance(contents, list):
                return json.dumps({"error": f"{path} is a file, not a directory"})

            # Process directory contents
            items = []
            for content in contents:
                item = {
                    "name": content.name,
                    "path": content.path,
                    "type": content.type,
                    "size": content.size,
                    "sha": content.sha,
                    "url": content.html_url,
                    "download_url": content.download_url,
                }
                items.append(item)

            # Sort by type (directories first) and then by name
            items.sort(key=lambda x: (x["type"] != "dir", x["name"].lower()))

            return json.dumps(items, indent=2)
        except GithubException as e:
            logger.error(f"Error getting directory contents: {e}")
            return json.dumps({"error": str(e)})

    def get_branch_content(self, repo_name: str, branch: str = "main") -> str:
        """Get the root directory content of a specific branch.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            branch (str, optional): The branch name. Defaults to "main".

        Returns:
            A JSON-formatted string containing a list of branch contents.
        """
        log_debug(f"Getting contents of branch {branch} in repository: {repo_name}")
        try:
            # This is just a convenience function that uses get_directory_content with empty path
            return self.get_directory_content(repo_name=repo_name, path="", ref=branch)
        except GithubException as e:
            logger.error(f"Error getting branch contents: {e}")
            return json.dumps({"error": str(e)})

    def create_branch(self, repo_name: str, branch_name: str, source_branch: Optional[str] = None) -> str:
        """Create a new branch in a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            branch_name (str): The name of the new branch.
            source_branch (str, optional): The source branch to create from. Defaults to repository's default branch.

        Returns:
            A JSON-formatted string containing information about the created branch.
        """
        log_debug(f"Creating branch {branch_name} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Get the source branch or default branch if not specified
            if source_branch is None:
                source_branch = repo.default_branch

            # Get the SHA of the latest commit on the source branch
            source_branch_ref = repo.get_git_ref(f"heads/{source_branch}")
            sha = source_branch_ref.object.sha

            # Create the new branch
            new_branch = repo.create_git_ref(f"refs/heads/{branch_name}", sha)

            branch_info = {
                "name": branch_name,
                "sha": new_branch.object.sha,
                "url": new_branch.url.replace("api.github.com/repos", "github.com").replace("git/refs/heads", "tree"),
            }

            return json.dumps(branch_info, indent=2)
        except GithubException as e:
            logger.error(f"Error creating branch: {e}")
            return json.dumps({"error": str(e)})

    def set_default_branch(self, repo_name: str, branch_name: str) -> str:
        """Set the default branch for a repository.

        Args:
            repo_name (str): The full name of the repository (e.g., 'owner/repo').
            branch_name (str): The name of the branch to set as default.

        Returns:
            A JSON-formatted string with success message or error.
        """
        log_debug(f"Setting default branch to {branch_name} in repository: {repo_name}")
        try:
            repo = self.g.get_repo(repo_name)

            # Check if the branch exists by looking at all branches
            branches = [branch.name for branch in repo.get_branches()]
            if branch_name not in branches:
                return json.dumps({"error": f"Branch '{branch_name}' does not exist"})

            # Set the default branch
            repo.edit(default_branch=branch_name)

            return json.dumps(
                {
                    "message": f"Default branch changed to {branch_name}",
                    "repository": repo_name,
                    "default_branch": branch_name,
                },
                indent=2,
            )
        except GithubException as e:
            logger.error(f"Error setting default branch: {e}")
            return json.dumps({"error": str(e)})

    def search_code(
        self,
        query: str,
        language: Optional[str] = None,
        repo: Optional[str] = None,
        user: Optional[str] = None,
        path: Optional[str] = None,
        filename: Optional[str] = None,
    ) -> str:
        """Search for code in GitHub repositories.

        Args:
            query (str): The search query.
            language (str, optional): Filter by language. Defaults to None.
            repo (str, optional): Filter by repository (e.g., 'owner/repo'). Defaults to None.
            user (str, optional): Filter by user or organization. Defaults to None.
            path (str, optional): Filter by file path. Defaults to None.
            filename (str, optional): Filter by filename. Defaults to None.

        Returns:
            A JSON-formatted string containing the search results.
        """
        log_debug(f"Searching code with query: {query}")
        try:
            search_query = query

            # Add filters to the query if provided
            if language:
                search_query += f" language:{language}"
            if repo:
                search_query += f" repo:{repo}"
            if user:
                search_query += f" user:{user}"
            if path:
                search_query += f" path:{path}"
            if filename:
                search_query += f" filename:{filename}"

            # Perform the search
            log_debug(f"Final search query: {search_query}")
            code_results = self.g.search_code(search_query)

            results: list[dict] = []
            limit = 60
            max_pages = 2  # GitHub returns 30 items per page, so 2 pages covers our limit
            page_index = 0

            while len(results) < limit and page_index < max_pages:
                # Fetch one page of results from GitHub API
                page_items = code_results.get_page(page_index)

                # Stop if no more results available
                if not page_items:
                    break

                # Process each code result in the current page
                for code in page_items:
                    code_info = {
                        "repository": code.repository.full_name,
                        "path": code.path,
                        "name": code.name,
                        "sha": code.sha,
                        "html_url": code.html_url,
                        "git_url": code.git_url,
                        "score": code.score,
                    }
                    results.append(code_info)
                page_index += 1

            # Return search results
            return json.dumps(
                {
                    "query": search_query,
                    "total_count": code_results.totalCount,
                    "results_count": len(results),
                    "results": results,
                },
                indent=2,
            )
        except GithubException as e:
            logger.error(f"Error searching code: {e}")
            return json.dumps({"error": str(e)})

    def search_issues_and_prs(
        self,
        query: str,
        state: Optional[str] = None,
        type_filter: Optional[str] = None,
        repo: Optional[str] = None,
        user: Optional[str] = None,
        label: Optional[str] = None,
        sort: str = "created",
        order: str = "desc",
        page: int = 1,
        per_page: int = 30,
    ) -> str:
        """Search for issues and pull requests on GitHub.

        Args:
            query (str): The search query.
            state (str, optional): Filter by state ('open', 'closed'). Defaults to None.
            type_filter (str, optional): Filter by type ('issue', 'pr'). Defaults to None.
            repo (str, optional): Filter by repository (e.g., 'owner/repo'). Defaults to None.
            user (str, optional): Filter by user or organization. Defaults to None.
            label (str, optional): Filter by label. Defaults to None.
            sort (str, optional): Sort results by ('created', 'updated', 'comments'). Defaults to "created".
            order (str, optional): Sort order ('asc', 'desc'). Defaults to "desc".
            page (int, optional): Page number for pagination. Defaults to 1.
            per_page (int, optional): Number of results per page. Defaults to 30.

        Returns:
            A JSON-formatted string containing the search results.
        """
        log_debug(f"Searching issues and PRs with query: {query}")
        try:
            search_query = query

            # Add filters to the query if provided
            if state:
                search_query += f" state:{state}"
            if type_filter == "issue":
                search_query += " is:issue"
            elif type_filter == "pr":
                search_query += " is:pr"
            if repo:
                search_query += f" repo:{repo}"
            if user:
                search_query += f" user:{user}"
            if label:
                search_query += f" label:{label}"

            # Perform the search
            log_debug(f"Final search query: {search_query}")
            issue_results = self.g.search_issues(search_query, sort=sort, order=order)

            # Process results
            per_page = min(per_page, 100)  # Ensure per_page doesn't exceed 100
            results = []

            try:
                # Get the specific page of results
                page_items = issue_results.get_page(page - 1)

                for issue in page_items:
                    issue_info = {
                        "number": issue.number,
                        "title": issue.title,
                        "repository": issue.repository.full_name,
                        "state": issue.state,
                        "created_at": issue.created_at.isoformat(),
                        "updated_at": issue.updated_at.isoformat(),
                        "html_url": issue.html_url,
                        "user": issue.user.login,
                        "is_pull_request": hasattr(issue, "pull_request") and issue.pull_request is not None,
                        "comments": issue.comments,
                        "labels": [label.name for label in issue.labels],
                    }
                    results.append(issue_info)

                    if len(results) >= per_page:
                        break
            except IndexError:
                # Page is out of range
                pass

            # Return search results
            return json.dumps(
                {
                    "query": search_query,
                    "total_count": issue_results.totalCount,
                    "page": page,
                    "per_page": per_page,
                    "results_count": len(results),
                    "results": results,
                },
                indent=2,
            )
        except GithubException as e:
            logger.error(f"Error searching issues and PRs: {e}")
            return json.dumps({"error": str(e)})



import json
from typing import Any, Dict, List, Union

from pydantic import BaseModel

from agno.models.message import Message


def serialize_input(
    input: Union[str, Dict[str, Any], List[Any], BaseModel, List[Message]],
) -> str:
    """Serialize the input to a string."""
    if isinstance(input, str):
        return input
    elif isinstance(input, dict):
        return json.dumps(input)
    elif isinstance(input, list):
        if any(isinstance(item, Message) for item in input):
            return json.dumps([item.to_dict() for item in input])
        else:
            return json.dumps(input)
    elif isinstance(input, BaseModel):
        return input.model_dump_json()



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from agno.os.routers.agents.schema import AgentResponse
from agno.os.routers.teams.schema import TeamResponse
from agno.workflow.agent import WorkflowAgent
from agno.workflow.workflow import Workflow


def _generate_schema_from_params(params: Dict[str, Any]) -> Dict[str, Any]:
    """Convert function parameters to JSON schema"""
    properties: Dict[str, Any] = {}
    required: List[str] = []

    for param_name, param_info in params.items():
        # Skip the default 'message' parameter for custom kwargs workflows
        if param_name == "message":
            continue

        # Map Python types to JSON schema types
        param_type = param_info.get("annotation", "str")
        default_value = param_info.get("default")
        is_required = param_info.get("required", False)

        # Convert Python type annotations to JSON schema types
        if param_type == "str":
            properties[param_name] = {"type": "string"}
        elif param_type == "bool":
            properties[param_name] = {"type": "boolean"}
        elif param_type == "int":
            properties[param_name] = {"type": "integer"}
        elif param_type == "float":
            properties[param_name] = {"type": "number"}
        elif "List" in str(param_type):
            properties[param_name] = {"type": "array", "items": {"type": "string"}}
        else:
            properties[param_name] = {"type": "string"}  # fallback

        # Add default value if present
        if default_value is not None:
            properties[param_name]["default"] = default_value

        # Add to required if no default value
        if is_required and default_value is None:
            required.append(param_name)

    schema = {"type": "object", "properties": properties}

    if required:
        schema["required"] = required

    return schema


def get_workflow_input_schema_dict(workflow: Workflow) -> Optional[Dict[str, Any]]:
    """Get input schema as dictionary for API responses"""

    # Priority 1: Explicit input_schema (Pydantic model)
    if workflow.input_schema is not None:
        try:
            return workflow.input_schema.model_json_schema()
        except Exception:
            return None

    # Priority 2: Auto-generate from custom kwargs
    if workflow.steps and callable(workflow.steps):
        custom_params = workflow.run_parameters
        if custom_params and len(custom_params) > 1:  # More than just 'message'
            return _generate_schema_from_params(custom_params)

    # Priority 3: No schema (expects string message)
    return None


class WorkflowResponse(BaseModel):
    id: Optional[str] = Field(None, description="Unique identifier for the workflow")
    name: Optional[str] = Field(None, description="Name of the workflow")
    db_id: Optional[str] = Field(None, description="Database identifier")
    description: Optional[str] = Field(None, description="Description of the workflow")
    input_schema: Optional[Dict[str, Any]] = Field(None, description="Input schema for the workflow")
    steps: Optional[List[Dict[str, Any]]] = Field(None, description="List of workflow steps")
    agent: Optional[AgentResponse] = Field(None, description="Agent configuration if used")
    team: Optional[TeamResponse] = Field(None, description="Team configuration if used")
    metadata: Optional[Dict[str, Any]] = Field(None, description="Additional metadata")
    workflow_agent: bool = Field(False, description="Whether this workflow uses a WorkflowAgent")
    is_component: bool = Field(False, description="Whether this workflow was created via Builder")
    current_version: Optional[int] = Field(None, description="Current published version number")
    stage: Optional[str] = Field(None, description="Stage of the loaded config (draft/published)")

    @classmethod
    async def _resolve_agents_and_teams_recursively(cls, steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Parse Agents and Teams into AgentResponse and TeamResponse objects.

        If the given steps have nested steps, recursively work on those."""
        if not steps:
            return steps

        def _prune_none(value: Any) -> Any:
            # Recursively remove None values from dicts and lists
            if isinstance(value, dict):
                return {k: _prune_none(v) for k, v in value.items() if v is not None}
            if isinstance(value, list):
                return [_prune_none(v) for v in value]
            return value

        for idx, step in enumerate(steps):
            if step.get("agent"):
                # Convert to dict and exclude fields that are None
                agent_response = await AgentResponse.from_agent(step.get("agent"))  # type: ignore
                step["agent"] = agent_response.model_dump(exclude_none=True)

            if step.get("team"):
                team_response = await TeamResponse.from_team(step.get("team"))  # type: ignore
                step["team"] = team_response.model_dump(exclude_none=True)

            if step.get("steps"):
                step["steps"] = await cls._resolve_agents_and_teams_recursively(step["steps"])

            # Prune None values in the entire step
            steps[idx] = _prune_none(step)

        return steps

    @classmethod
    async def from_workflow(
        cls,
        workflow: Workflow,
        is_component: bool = False,
    ) -> "WorkflowResponse":
        workflow_dict = workflow.to_dict_for_steps()
        steps = workflow_dict.get("steps")

        if steps:
            steps = await cls._resolve_agents_and_teams_recursively(steps)

        return cls(
            id=workflow.id,
            name=workflow.name,
            db_id=workflow.db.id if workflow.db else None,
            description=workflow.description,
            steps=steps,
            input_schema=get_workflow_input_schema_dict(workflow),
            metadata=workflow.metadata,
            workflow_agent=isinstance(workflow.agent, WorkflowAgent) if workflow.agent else False,
            is_component=is_component,
            current_version=getattr(workflow, "_version", None),
            stage=getattr(workflow, "_stage", None),
        )
